@@ -4,7 +4,7 @@
 #include <stdio.h>
 #include <string>
 
-// A stdio FILE wrapper
+// A stdio FILE wrapper(with some BSD and POSIXisms, and a little dash of Win32, thrown in for special behaviors)
 class FileWrapper
 {
  public:
@@ -16,7 +16,7 @@ class FileWrapper
   MODE_WRITE_SAFE	// Will throw an exception instead of overwriting an existing file.
  };
 
- FileWrapper(const char *path, const int mode);
+ FileWrapper(const char *path, const int mode, const char *purpose = NULL);
  #if 0
  FileWrapper();
  #endif
@@ -46,6 +46,14 @@ class FileWrapper
  void flush(void);
  //void flushsync(void);	// TODO: see fflush and fsync
 
+ void close(void);	// Flushes and closes the underlying OS/C lib file.  Calling any other method of this class after a call to
+			// this method is illegal(except for the implicit call to the destructor).
+			//
+			// This is necessary since there can be errors when closing a file, and we can't safely throw an
+			// exception from the destructor.
+			//
+			// Manually calling this method isn't strictly necessary, it'll be called from the destructor
+			// automatically, but calling is strongly recommended when the file is opened for writing.
  private:
 
  FileWrapper & operator=(const FileWrapper &);    // Assignment operator
