@@ -425,6 +425,8 @@ bool PCECD_SetSettings(const PCECD_Settings *settings)
         ADPCMLP = settings ? settings->ADPCM_LPF : 0;
 
 	SCSICD_SetTransferRate(126000 * (settings ? settings->CD_Speed : 1));
+
+	return true;
 }
 
 bool PCECD_Init(const PCECD_Settings *settings, void (*irqcb)(bool), double master_clock, unsigned int ocm, Blip_Buffer *soundbuf_l, Blip_Buffer *soundbuf_r)
@@ -461,12 +463,13 @@ void PCECD_Close(void)
          MDFN_free(ADPCM.RAM);
          ADPCM.RAM = NULL;
         }
+	SCSICD_Close();
 }
 
 
 int32 PCECD_Power(uint32 timestamp)
 {
-	if(timestamp != lastts)
+	if((int32)timestamp != lastts)
  	 (void)PCECD_Run(timestamp);
 
 	IRQCB(0);
