@@ -34,6 +34,9 @@
 #include "sound.h"
 #include "z80.h"
 
+namespace MDFN_IEN_GB
+{
+
 static uint32 *gbColorFilter = NULL; //[32768];
 static uint32 gbMonoColorMap[8 + 1];	// Mono color map(+1 = LCD off color)!
 
@@ -128,30 +131,32 @@ uint8 register_SVBK  = 0;
 uint8 register_IE    = 0;
 
 // ticks definition
-int GBDIV_CLOCK_TICKS          = 64;
-int GBLCD_MODE_0_CLOCK_TICKS   = 51;
-int GBLCD_MODE_1_CLOCK_TICKS   = 1140;
-int GBLCD_MODE_2_CLOCK_TICKS   = 20;
-int GBLCD_MODE_3_CLOCK_TICKS   = 43;
-int GBLY_INCREMENT_CLOCK_TICKS = 114;
-int GBTIMER_MODE_0_CLOCK_TICKS = 256;
-int GBTIMER_MODE_1_CLOCK_TICKS = 4;
-int GBTIMER_MODE_2_CLOCK_TICKS = 16;
-int GBTIMER_MODE_3_CLOCK_TICKS = 64;
-int GBSERIAL_CLOCK_TICKS       = 128;
-int GBSYNCHRONIZE_CLOCK_TICKS  = 52920;
+static int GBDIV_CLOCK_TICKS          = 64;
+static int GBLCD_MODE_0_CLOCK_TICKS   = 51;
+static int GBLCD_MODE_1_CLOCK_TICKS   = 1140;
+static int GBLCD_MODE_2_CLOCK_TICKS   = 20;
+static int GBLCD_MODE_3_CLOCK_TICKS   = 43;
+static int GBLY_INCREMENT_CLOCK_TICKS = 114;
+static int GBTIMER_MODE_0_CLOCK_TICKS = 256;
+static int GBTIMER_MODE_1_CLOCK_TICKS = 4;
+static int GBTIMER_MODE_2_CLOCK_TICKS = 16;
+static int GBTIMER_MODE_3_CLOCK_TICKS = 64;
+static int GBSERIAL_CLOCK_TICKS       = 128;
+static int GBSYNCHRONIZE_CLOCK_TICKS  = 52920;
 
 // state variables
+static int32 snooze;
+static int32 PadInterruptDelay;
 
 // serial
-int gbSerialOn = 0;
-int gbSerialTicks = 0;
-int gbSerialBits = 0;
+static int gbSerialOn;
+static int gbSerialTicks;
+static int gbSerialBits;
 // timer
-int gbTimerOn = 0;
-int gbTimerTicks = GBTIMER_MODE_0_CLOCK_TICKS;
-int gbTimerClockTicks = GBTIMER_MODE_0_CLOCK_TICKS;
-int gbTimerMode = 0;
+static int gbTimerOn;
+static int gbTimerTicks;
+static int gbTimerClockTicks;
+static int gbTimerMode;
 
 
 enum
@@ -312,6 +317,17 @@ static bool LoadCPalette(const char *syspalname, uint8 **ptr, uint32 num_entries
   return(false);
  }
 
+ // Print a warning message about unused trailing data
+ {
+  int64 rs;
+
+  fseek(fp, 0, SEEK_END);
+  if((rs = ftell(fp)) > (num_entries * 3))
+  {
+   MDFN_printf(_("Warning: %lld byte(s) of trailing unused data.\n"), (long long)(rs - (num_entries * 3)));
+  }
+ }
+
  fclose(fp);
 
  MDFN_indent(-1);
@@ -1257,6 +1273,9 @@ void gbReset()
 
 static void gbPower(void)
 {
+ snooze = 0;
+ PadInterruptDelay = 0;
+
   if(gbCgbMode)
   {
    memset(gbWram,0,0x8000);
@@ -1723,6 +1742,10 @@ static SFORMAT gbSaveGameStruct[] =
   SFVAR(GBTIMER_MODE_3_CLOCK_TICKS),
   SFVAR(GBSERIAL_CLOCK_TICKS),
   SFVAR(GBSYNCHRONIZE_CLOCK_TICKS),
+
+  SFVAR(snooze),
+  SFVAR(PadInterruptDelay),
+
   SFVAR(gbDivTicks),
   SFVAR(gbLcdMode),
   SFVAR(gbLcdTicks),
@@ -2055,6 +2078,8 @@ static int Load(const char *name, MDFNFILE *fp)
   return(0);
  }
 
+ gbEmulatorType = MDFN_GetSettingI("gb.system_type");
+
  gbRom = (uint8 *)malloc(fp->size);
  memcpy(gbRom, fp->data, fp->size);
  gbRomSize = fp->size;
@@ -2086,12 +2111,16 @@ static int Load(const char *name, MDFNFILE *fp)
  gblayerSettings = 0xFF;
 
  // Custom palettes
- if(!LoadCPalette("gb", &Custom_GB_ColorMap, 4))
+ if(gbCgbMode)
+ {
+  if(!LoadCPalette("gbc", &Custom_GBC_ColorMap, 32768))
   return(0);
-
- if(!LoadCPalette("gbc", &Custom_GBC_ColorMap, 32768))
-  return(0);
-
+ }
+ else
+ {
+  if(!LoadCPalette("gb", &Custom_GB_ColorMap, 4))
+   return(0);
+ }
 
  return(1);
 }
@@ -2267,9 +2296,6 @@ static void MDFNGB_SetInput(int port, const char *type, void *ptr)
  paddie = (uint8*)ptr;
 }
 
-static int32 snooze = 0;
-static int32 PadInterruptDelay = 0;
-
 static void Emulate(EmulateSpecStruct *espec)
 {
  bool linedrawn[144];
@@ -2635,8 +2661,18 @@ static void DoSimpleCommand(int cmd)
  }
 }
 
+static const MDFNSetting_EnumList SystemType_List[] =
+{
+ { "auto", 0, gettext_noop("Auto"), gettext_noop("Automatic detection based on headers.") },
+ { "dmg", 3, gettext_noop("DMG"), gettext_noop("Original GameBoy Monochrome.") },
+ { "cgb", 1, gettext_noop("CGB"), gettext_noop("GameBoy Color.\n\nThis option is not fully implemented in regards to handling of DMG games.") },
+ { "agb", 4, gettext_noop("AGB"), gettext_noop("GameBoy Advance.\n\nThis option is not fully implemented in regards to handling of DMG games.") },
+ { NULL, 0 },
+};
+
 static MDFNSetting GBSettings[] =
 {
+ { "gb.system_type", MDFNSF_EMU_STATE | MDFNSF_UNTRUSTED_SAFE, gettext_noop("Emulated GB type."), NULL, MDFNST_ENUM, "auto", NULL, NULL, NULL, NULL, SystemType_List },
  { NULL }
 };
 
@@ -2689,6 +2725,10 @@ static const FileExtensionSpecStruct KnownExtensions[] =
  { NULL, NULL }
 };
 
+}
+
+using namespace MDFN_IEN_GB;
+
 MDFNGI EmulatedGB =
 {
  "gb",
