@@ -43,9 +43,11 @@
 #include <trio/trio.h>
 #include <errno.h>
 
-
 #include "../PSFLoader.h"
 
+namespace MDFN_IEN_GBA
+{
+
 class GSFLoader : public PSFLoader
 {
  public:
@@ -3289,6 +3291,9 @@ static void Emulate(EmulateSpecStruct *espec)
  while(!frameready && (soundTS < 300000))
   CPULoop(espec->surface, 300000);
 
+ if(GBA_RTC)
+  GBA_RTC->AddTime(soundTS);
+
  espec->MasterCycles = soundTS;
 
  espec->SoundBufSize = MDFNGBASOUND_Flush(espec->SoundBuf, espec->SoundBufMaxSize);
@@ -3372,12 +3377,18 @@ static InputInfoStruct InputInfo =
 
 static const FileExtensionSpecStruct KnownExtensions[] =
 {
+ { ".gsf", gettext_noop("GSF Rip") },
+ { ".minigsf", gettext_noop("MiniGSF Rip") },
  { ".gba", gettext_noop("GameBoy Advance ROM Image") },
  { ".agb", gettext_noop("GameBoy Advance ROM Image") },
  { ".bin", gettext_noop("GameBoy Advance ROM Image") },
  { NULL, NULL }
 };
 
+}
+
+using namespace MDFN_IEN_GBA;
+
 MDFNGI EmulatedGBA =
 {
  "gba",
