@@ -36,4 +36,6 @@ std::string MDFN_MakeFName(MakeFName_Type type, int id1, const char *cd1);
 
 const char * GetFNComponent(const char *str);
 
+void MDFN_GetFilePathComponents(const std::string &file_path, std::string *dir_path_out, std::string *file_base_out = NULL, std::string *file_ext_out = NULL);
+std::string MDFN_EvalFIP(const std::string &dir_path, const std::string &rel_path, bool skip_safety_check = false);
 #endif
