@@ -11,6 +11,13 @@
 namespace MDFN
 {
 
+class LEPacker;
+class LEPackable
+{
+ public:
+ virtual void pack(LEPacker &lep) = 0;
+};
+
 class LEPacker : public std::vector<uint8>
 {
  public:
@@ -31,6 +38,11 @@ class LEPacker : public std::vector<uint8>
   read_pos = 0;
  }
 
+ void operator^(LEPackable &o)
+ {
+  o.pack(*this);
+ }
+
  template<typename T> void operator^(T &val)
  {
   size_type csize = size();
