@@ -6,8 +6,8 @@ extern c68k_struc Sub68K;
 void MDCD_Run(int32 md_master_cycles);
 void MDCD_PCM_Run(int32 cycles);
 void MDCD_Reset(void);
-bool MDCD_Load(md_game_info *);
-bool MDCD_TestMagic(void);
+bool MDCD_Load(std::vector<CDIF *> *CDInterfaces, md_game_info *);
+bool MDCD_TestMagic(std::vector<CDIF *> *CDInterfaces);
 void MDCD_Close(void);
 
 #endif
