@@ -160,11 +160,11 @@ void MDVDP::Reset(void)
     dma_fill = 0;
     im2_flag = 0;
 
-    vdp_cycle_counter = 0;
+    vdp_cycle_counter = 4 * 10;
     vdp_last_ts = 0;
-    vdp_line_phase = 0;
+    vdp_line_phase = VDPLP_VISIBLE_0;
     vdp_hcounter_start_ts = 0;
-    scanline = 0;
+    scanline = visible_frame_end;
 
     MD_Suspend68K(FALSE);
     z80_set_interrupt(FALSE);
@@ -839,11 +839,6 @@ void MDVDP::SetSurface(EmulateSpecStruct *espec_arg)	//MDFN_Surface *new_surface
  espec = espec_arg;
  surface = espec->surface;
  rect = &espec->DisplayRect;
-
- //rect->x = bitmap.viewport.x;
- //rect->y = 0; //bitmap.viewport.y;
- //rect->w = bitmap.viewport.w;
- //rect->h = bitmap.viewport.h;
 }
 
 
@@ -908,22 +903,16 @@ void MDVDP::render_line(int line)
     // Our display output window is nominally XXX*224 with NTSC, XXX*240 with PAL.
     //if(line == 0)
     {
-     rect->x = 0;
-     rect->y = is_pal ? 0 : 8;
-     rect->w = WantAutoAspect ? vp_w : 320;
-     rect->h = is_pal ? 240 : 224;
-
      if(!(reg[0x1] & 0x8) && is_pal)
      {
       //puts("FIXME, lines clear");
      }
     }
 
-    // Double-vertical-resolution interlaced mode
-    if((reg[0xC] & 0x06) == 0x06)
+    if(line == 120)
     {
-     rect->y *= 2;
-     rect->h *= 2;
+     rect->x = 0;
+     rect->w = WantAutoAspect ? vp_w : 320;
     }
 
     if((reg[1] & 0x40) == 0x00)
