@@ -16,7 +16,7 @@
  */
 
 /* This resampler has only been designed with NES CPU frequencies(NTSC and PAL) as the input rate, and output rates of
-   22050-48000 in mind, up to 1024 coefficient multiply-accumulates per output sample.
+   22050-1048576 in mind, up to 1024 coefficient multiply-accumulates per output sample.
 */
 
 // Don't set these higher than 3, the accumulation variables will overflow if you do.
@@ -324,6 +324,7 @@ int32 NES_Resampler::Do(int16 *in, int16 *out, uint32 maxoutlen, uint32 inlen, i
 {
 	uint32 max;
 	uint32 count=0;
+	int32 *boobuf = &IntermediateBuffer[0];
 	int32 *I32Out = boobuf;
 
 	{
@@ -505,6 +506,8 @@ NES_Resampler::NES_Resampler(double input_rate, double output_rate, double rate_
  DebiasCorner = debias_corner;
  Quality = quality;
 
+ IntermediateBuffer.resize(OutputRate * 4 / 50);	// *4 for safety padding, / min(50,60), an approximate calculation
+
  cpuext = cputest_get_flags();
 
  MDFN_printf("filter.cpp debug info:\n");
@@ -623,7 +626,26 @@ NES_Resampler::NES_Resampler(double input_rate, double output_rate, double rate_
   MDFN_printf("Desired rate error: %.10f, Actual rate error: %.10f\n", rate_error, fabs((double)input_rate / output_rate * ratio - 1));
  }
 
- cutoff = (ratio - required_bandwidth) / NumPhases;
+ // Optimize cutoff for maximum sound frequencies humans can perceive, keep individual coefficient values from
+ // becoming too large, and keep dogs from freaking out with some games that "silence"
+ // sounds by making their fundamental frequencies beyond the range of typical human hearing.
+ //
+ // (For when output playback rate is considerably higher than 48KHz)
+ //
+ // This also incidentally handles the case of UPSAMPLING to a higher rate(at least when upsampling from the NES
+ // CPU frequency), so if this code is removed/disabled, code to handle that case will need to be added.
+ {
+  double cutoff_robot = (ratio - required_bandwidth) / NumPhases;
+  double cutoff_human = ((24000 * 2) / input_rate) / NumPhases; // Not subtracting required_bandwidth here is intentional.
+
+  if(cutoff_human < cutoff_robot)
+  {
+   //puts("HUMANS WIN THIS ROUND");
+   cutoff = cutoff_human;
+  }
+  else
+   cutoff = cutoff_robot;
+ }
 
  MDFN_printf("Cutoff: %f, %f\n", cutoff, required_bandwidth);
 
