@@ -607,18 +607,12 @@ static void RefreshLine(int lastpixel)
    {
     const int MMC5Mode = 4;
 
-    ys=((scanline>>3)+MMC5HackSPScroll)&0x1F;
-    if(ys>=0x1E) ys-=0x1E;
-
     #include "ppu-subline.h"
    }
    else if (MMC5HackCHRMode==1 && (MMC5HackSPMode&0x80))
    {
     const int MMC5Mode = 3;
 
-    ys=((scanline>>3)+MMC5HackSPScroll)&0x1F;
-    if(ys>=0x1E) ys-=0x1E;
-
     #include "ppu-subline.h"
    }
    else if(MMC5HackCHRMode == 1)
@@ -767,6 +761,12 @@ static void DoLine(MDFN_Surface *surface, int skip)
 
  if(scanline >= 0 && MMC5Hack && (ScreenON || SpriteON)) MMC5_hb(scanline);
 
+ if(MMC5Hack)
+ {
+  ys=((scanline>>3)+MMC5HackSPScroll)&0x1F;
+  if(ys>=0x1E) ys-=0x1E;
+ }
+
  X6502_Run(256);
  if(firstpixel < 240)
  {
@@ -1310,6 +1310,16 @@ void NESPPU_GetDisplayRect(MDFN_Rect *DisplayRect)
  memcpy(DisplayRect, &PPUDisplayRect, sizeof(MDFN_Rect));
 }
 
+void NESPPU_TranslateMouseXY(uint32 &new_x, uint32 &new_y)
+{
+ MDFN_Rect dr;
+
+ NESPPU_GetDisplayRect(&dr);
+
+ new_x = (uint32)(((new_x + ((double)dr.x * MDFNGameInfo->nominal_width) / dr.w) * 256 / MDFNGameInfo->nominal_width));
+ new_y = new_y + dr.y;
+}
+
 static void RedoRL(void)
 {
  PPUDisplayRect.x = NTSCBlitter ? 4 : 0;
@@ -1349,6 +1359,15 @@ static void RedoRL(void)
  MDFNGameInfo->fb_width = (NTSCBlitter ? 768 : 256);
 }
 
+void MDFNPPU_Close(void)
+{
+ if(NTSCBlitter)
+ {
+  free(NTSCBlitter);
+  NTSCBlitter = NULL;
+ }
+}
+
 void MDFNPPU_Init(void) 
 {
  makeppulut();
