@@ -26,6 +26,7 @@
 
 #define SQ_SHIFT        8
 #define TRINPCM_SHIFT   0
+#include <math.h>
 
 static void DoSQ1(void);
 static void DoSQ2(void);
@@ -777,9 +778,17 @@ int FlushEmulateSound(int reverse, int16 *SoundBuf, int32 MaxSoundFrames)
 
      #if 0
      {
-      static uint32 meow = 0;
-      *tmpo = (((meow & 32) * 65535) / 32) - 32768; //(meow & 32) * 29200 / 32;
-      meow++;
+static double phase = 0;
+static double phase_inc = 0.000;
+static double phase_inc_inc = 0.000000001;
+
+ *tmpo = 16384 * sin(phase);
+ phase += phase_inc;
+ phase_inc += phase_inc_inc;
+
+      //static uint32 meow = 0;
+      //*tmpo = (((meow & 32) * 65535) / 32) - 32768; //(meow & 32) * 29200 / 32;
+      //meow++;
       // *tmpo = (rand() & 0x7FFF) * 28000 / 32768;
       //*tmpo = (int16)(rand() & 0xFFFF); // * 28000 * 2 / 32768;
      }
