@@ -22,6 +22,7 @@
 #include "huc.h"
 #include "subhw.h"
 #include "../cdrom/pcecd.h"
+#include "../cdrom/scsicd.h"
 #include "hes.h"
 #include "debug.h"
 #include "tsushin.h"
@@ -47,6 +48,10 @@ static const MDFNSetting_EnumList PSGRevisionList[] =
  { NULL, 0 },
 };
 
+static std::vector<CDIF*> *cdifs = NULL;
+static bool CD_TrayOpen;
+static int CD_SelectedDisc;	// -1 for no disc
+
 HuC6280 *HuCPU;
 
 VCE *vce = NULL;
@@ -55,6 +60,7 @@ static PCE_PSG *psg = NULL;
 
 extern ArcadeCard *arcade_card;	// Bah, lousy globals.
 
+Blip_Buffer huc_sbuf;
 static Blip_Buffer sbuf[2];
 static bool SetSoundRate(double rate);
 
@@ -382,8 +388,6 @@ static int Load(const char *name, MDFNFILE *fp)
   if(!PCE_HESLoad(fp->data, fp->size))
    return(0);
 
-  // FIXME:  If a HES rip tries to execute a SCSI command, the CD emulation code will probably crash.  Obviously, a HES rip shouldn't do this,
-  // but Mednafen shouldn't crash either. ;)
   PCE_IsCD = 1;
   PCECD_Init(NULL, PCECDIRQCB, PCE_MASTER_CLOCK, 1, &sbuf[0], &sbuf[1]);
  }
@@ -436,6 +440,12 @@ static int Load(const char *name, MDFNFILE *fp)
 
 static void LoadCommonPre(void)
 {
+ // Initialize sound buffers
+ for(unsigned i = 0; i < 2; i++)
+  sbuf[i].clear();
+
+ huc_sbuf.clear();
+
  // FIXME:  Make these globals less global!
  PCE_ACEnabled = MDFN_GetSettingB("pce.arcadecard");
 
@@ -565,25 +575,25 @@ static int LoadCommon(void)
  return(1);
 }
 
-static bool DetectGECD(void)	// Very half-assed detection until(if) we get ISO-9660 reading code.
+static bool DetectGECD(CDIF *cdiface)	// Very half-assed detection until(if) we get ISO-9660 reading code.
 {
  uint8 sector_buffer[2048];
- CD_TOC toc;
+ CDUtility::TOC toc;
 
- CDIF_ReadTOC(&toc);
+ cdiface->ReadTOC(&toc);
 
  // Now, test for the Games Express CD games.  The GE BIOS seems to always look at sector 0x10, but only if the first track is a
  // data track.
  if(toc.first_track == 1 && (toc.tracks[1].control & 0x4))
  {
-  if(CDIF_ReadSector(sector_buffer, 0x10, 1) == 0x1)
+  if(cdiface->ReadSector(sector_buffer, 0x10, 1) == 0x1)
   {
    if(!memcmp((char *)sector_buffer + 0x8, "HACKER CD ROM SYSTEM", 0x14))
     return(true);
 
    if(!memcmp((char *)sector_buffer + 0x01, "CD001", 0x5))
    {
-    if(CDIF_ReadSector(sector_buffer, 0x14, 1) == 0x1)
+    if(cdiface->ReadSector(sector_buffer, 0x14, 1) == 0x1)
     {
      static const uint32 known_crcs[] =
      {
@@ -606,24 +616,25 @@ static bool DetectGECD(void)	// Very half-assed detection until(if) we get ISO-9
  return(false);
 }
 
-static bool TestMagicCD(void)
+static bool TestMagicCD(std::vector<CDIF *> *CDInterfaces)
 {
  static const uint8 magic_test[0x20] = { 0x82, 0xB1, 0x82, 0xCC, 0x83, 0x76, 0x83, 0x8D, 0x83, 0x4F, 0x83, 0x89, 0x83, 0x80, 0x82, 0xCC,  
 				   	 0x92, 0x98, 0x8D, 0xEC, 0x8C, 0xA0, 0x82, 0xCD, 0x8A, 0x94, 0x8E, 0xAE, 0x89, 0xEF, 0x8E, 0xD0
 				       };
+ CDIF *cdiface = (*CDInterfaces)[0];
  uint8 sector_buffer[2048];
- CD_TOC toc;
+ CDUtility::TOC toc;
  bool ret = FALSE;
 
  memset(sector_buffer, 0, sizeof(sector_buffer));
 
- CDIF_ReadTOC(&toc);
+ cdiface->ReadTOC(&toc);
 
  for(int32 track = toc.first_track; track <= toc.last_track; track++)
  {
   if(toc.tracks[track].control & 0x4)
   {
-   if(CDIF_ReadSector(sector_buffer, toc.tracks[track].lba, 1) != 0x1)
+   if(cdiface->ReadSector(sector_buffer, toc.tracks[track].lba, 1) != 0x1)
     break;
 
    if(!memcmp((char*)sector_buffer, (char *)magic_test, 0x20))
@@ -640,7 +651,7 @@ static bool TestMagicCD(void)
  {
   if(toc.tracks[track].control & 0x4)
   {
-   CDIF_ReadSector(sector_buffer, toc.tracks[track].lba, 1);
+   cdiface->ReadSector(sector_buffer, toc.tracks[track].lba, 1);
    if(!strncmp("PC-FX:Hu_CD-ROM", (char*)sector_buffer, strlen("PC-FX:Hu_CD-ROM")))
    {
     return(false);
@@ -648,13 +659,13 @@ static bool TestMagicCD(void)
   }
  }
 
- if(DetectGECD())
+ if(DetectGECD(cdiface))
   ret = true;
 
  return(ret);
 }
 
-static int LoadCD(void)
+static int LoadCD(std::vector<CDIF *> *CDInterfaces)
 {
  static const FileExtensionSpecStruct KnownBIOSExtensions[] =
  {
@@ -673,7 +684,7 @@ static int LoadCD(void)
 
  LoadCommonPre();
 
- const char *bios_sname = DetectGECD() ? "pce.gecdbios" : "pce.cdbios";
+ const char *bios_sname = DetectGECD((*CDInterfaces)[0]) ? "pce.gecdbios" : "pce.cdbios";
  std::string bios_path = MDFN_MakeFName(MDFNMKF_FIRMWARE, 0, MDFN_GetSettingS(bios_sname).c_str() );
 
  if(!fp.Open(bios_path, KnownBIOSExtensions, _("CD BIOS")))
@@ -708,6 +719,15 @@ static int LoadCD(void)
   return(0);
  }
 
+ MDFNGameInfo->GameType = GMT_CDROM;
+ CD_TrayOpen = false;
+ CD_SelectedDisc = 0;
+ cdifs = CDInterfaces;
+
+ SCSICD_SetDisc(true, NULL, true);
+ SCSICD_SetDisc(false, (*CDInterfaces)[0], true);
+
+
  MDFN_printf(_("CD Layout:   0x%s\n"), md5_context::asciistr(MDFNGameInfo->MD5, 0).c_str());
  MDFN_printf(_("Arcade Card Emulation:  %s\n"), PCE_ACEnabled ? _("Enabled") : _("Disabled"));
 
@@ -782,17 +802,30 @@ static void Emulate(EmulateSpecStruct *espec)
 
   psg->EndFrame(HuCPU->Timestamp() / 3);
 
+  HuC_EndFrame(HuCPU->Timestamp());
+
   //assert(!(HuCPU->Timestamp() % 3));
 
   if(espec->SoundBuf)
   {
    int32 new_sc;
+
+   huc_sbuf.end_frame(HuCPU->Timestamp() / 3);
+
    for(int y = 0; y < 2; y++)
    {
     sbuf[y].end_frame(HuCPU->Timestamp() / 3);
 
     new_sc = sbuf[y].read_samples(espec->SoundBuf + espec->SoundBufSize * 2 + y, espec->SoundBufMaxSize - espec->SoundBufSize, 1);
    }
+
+   if(huc_sbuf.clear_modified())
+   {
+
+   }
+   else
+    huc_sbuf.remove_samples(huc_sbuf.samples_avail());
+
    espec->SoundBufSize += new_sc;
   }
 
@@ -842,6 +875,10 @@ static int StateAction(StateMem *sm, int load, int data_only)
  {
   SFARRAY(BaseRAM, IsSGX? 32768 : 8192),
   SFVAR(PCE_TimestampBase),
+
+  SFVAR(CD_TrayOpen),
+  SFVAR(CD_SelectedDisc),
+
   SFEND
  };
 
@@ -857,7 +894,14 @@ static int StateAction(StateMem *sm, int load, int data_only)
 
  if(load)
  {
+  if(cdifs)
+  {
+   // Sanity check.
+   if(CD_SelectedDisc >= (int)cdifs->size())
+    CD_SelectedDisc = (int)cdifs->size() - 1;
 
+   SCSICD_SetDisc(CD_TrayOpen, (CD_SelectedDisc >= 0 && !CD_TrayOpen) ? (*cdifs)[CD_SelectedDisc] : NULL, true);
+  }
  }
 
  return(ret);
@@ -889,12 +933,67 @@ void PCE_Power(void)
  }
 }
 
+static void CDInsertEject(void)
+{
+ CD_TrayOpen = !CD_TrayOpen;
+
+ for(unsigned disc = 0; disc < cdifs->size(); disc++)
+ {
+  if(!(*cdifs)[disc]->Eject(CD_TrayOpen))
+  {
+   MDFN_DispMessage(_("Eject error."));
+   CD_TrayOpen = !CD_TrayOpen;
+  }
+ }
+
+ if(CD_TrayOpen)
+  MDFN_DispMessage(_("Virtual CD Drive Tray Open"));
+ else
+  MDFN_DispMessage(_("Virtual CD Drive Tray Closed"));
+
+ SCSICD_SetDisc(CD_TrayOpen, (CD_SelectedDisc >= 0 && !CD_TrayOpen) ? (*cdifs)[CD_SelectedDisc] : NULL);
+}
+
+static void CDEject(void)
+{
+ if(!CD_TrayOpen)
+  CDInsertEject();
+}
+
+static void CDSelect(void)
+{
+ if(cdifs && CD_TrayOpen)
+ {
+  CD_SelectedDisc = (CD_SelectedDisc + 1) % (cdifs->size() + 1);
+
+  if((unsigned)CD_SelectedDisc == cdifs->size())
+   CD_SelectedDisc = -1;
+
+  if(CD_SelectedDisc == -1)
+   MDFN_DispMessage(_("Disc absence selected."));
+  else
+   MDFN_DispMessage(_("Disc %d of %d selected."), CD_SelectedDisc + 1, (int)cdifs->size());
+ }
+}
+
 static void DoSimpleCommand(int cmd)
 {
  switch(cmd)
  {
   case MDFN_MSC_RESET: PCE_Power(); break;
   case MDFN_MSC_POWER: PCE_Power(); break;
+
+  case MDFN_MSC_INSERT_DISK:
+	CDInsertEject();
+        break;
+
+  case MDFN_MSC_SELECT_DISK:
+	CDSelect();
+        break;
+
+  case MDFN_MSC_EJECT_DISK:
+	CDEject();
+        break;
  }
 }
 
@@ -989,6 +1088,10 @@ static const FileExtensionSpecStruct KnownExtensions[] =
 
 static bool SetSoundRate(double rate)
 {
+ huc_sbuf.set_sample_rate(rate ? rate : 44100, 50);
+ huc_sbuf.clock_rate((long)(PCE_MASTER_CLOCK / 3));
+ huc_sbuf.bass_freq(20);
+
  for(int y = 0; y < 2; y++)
  {
   sbuf[y].set_sample_rate(rate ? rate : 44100, 50);
