@@ -1180,7 +1180,7 @@ void VDC_RunFrame(MDFN_Surface *surface, MDFN_Rect *DisplayRect, MDFN_Rect *Line
 
  do
  {
-  const bool SHOULD_DRAW = (!skip && frame_counter >= (DisplayRect->y + 14) && frame_counter < (DisplayRect->y + DisplayRect->h + 14));
+  const bool SHOULD_DRAW = (!skip && (int)frame_counter >= (DisplayRect->y + 14) && (int)frame_counter < (DisplayRect->y + DisplayRect->h + 14));
 
   vdc = vdc_chips[0];
 
@@ -1477,7 +1477,9 @@ void VDC_RunFrame(MDFN_Surface *surface, MDFN_Rect *DisplayRect, MDFN_Rect *Line
 
   if(PCE_IsCD)
   {
-   PCECD_Run(HuCPU.timestamp * 3);
+   int32 dummy_ne;
+
+   dummy_ne = PCECD_Run(HuCPU.timestamp * 3);
   }
   for(int chip = 0; chip < VDC_TotalChips; chip++)
   {
