@@ -33,6 +33,8 @@ PS_CPU::PS_CPU()
 {
  uint32 dp_offset;
 
+ Halted = false;
+
  memset(FastMap, 0, sizeof(FastMap));
  memset(DummyPage, 0x00, sizeof(DummyPage));
 
@@ -73,8 +75,16 @@ INLINE void PS_CPU::RecalcIPCache(void)
 
  if((CP0.SR & CP0.CAUSE & 0xFF00) && (CP0.SR & 1))
   IPCache = 0x80;
+
+ if(Halted)
+  IPCache = 0x80;
 }
 
+void PS_CPU::SetHalt(bool status)
+{
+ Halted = status;
+ RecalcIPCache();
+}
 
 void PS_CPU::Power(void)
 {
@@ -86,11 +96,16 @@ void PS_CPU::Power(void)
  gte_ts_done = 0;
 
  BACKED_PC = 0xBFC00000;
- BACKED_new_PC = BACKED_PC + 4;
+ BACKED_new_PC = 4;
+ BACKED_new_PC_mask = ~0U;
  InBDSlot = false;
 
  LDWhich = 0x20;
  LDValue = 0;
+ LDPenalty = 0;
+
+ ReadPenaltyCounter = 0;
+ ReadPenaltyMask = 0;
  
  CP0.SR |= (1 << 22);	// BEV
  CP0.SR |= (1 << 21);	// TS
@@ -124,7 +139,8 @@ INLINE T PS_CPU::ReadMemory(pscpu_timestamp_t &timestamp, uint32 address, bool D
 {
  T ret;
 
- timestamp++;
+ timestamp += 2;
+ //timestamp++;
  //assert(!(CP0.SR & 0x10000));
 
  if(sizeof(T) == 1)
@@ -171,7 +187,8 @@ uint32 PS_CPU::Exception(uint32 code, uint32 PC)
 
  if(code != EXCEPTION_INT && code != EXCEPTION_BP && code != EXCEPTION_SYSCALL)
  {
-  printf("Exception: %08x @ PC=0x%08x\n", code, PC);
+  printf("Exception: %08x @ PC=0x%08x %d -- IPCache=0x%02x -- IPEND=0x%02x -- SR=0x%08x ; IRQC_Status=0x%04x -- IRQC_Mask=0x%04x\n", code, PC, InBDSlot, IPCache, (CP0.CAUSE >> 8) & 0xFF, CP0.SR,
+	IRQ_GetRegister(IRQ_GSREG_STATUS, NULL, 0), IRQ_GetRegister(IRQ_GSREG_MASK, NULL, 0));
   //assert(0);
  }
 
@@ -203,14 +220,13 @@ uint32 PS_CPU::Exception(uint32 code, uint32 PC)
  return(handler);
 }
 
-extern int32 timestamp_funny;
-
 template<bool DebugMode>
 pscpu_timestamp_t PS_CPU::RunReal(pscpu_timestamp_t timestamp_in)
 {
  register pscpu_timestamp_t timestamp = timestamp_in;
  register uint32 PC = BACKED_PC;
  register uint32 new_PC = BACKED_new_PC;
+ register uint32 new_PC_mask = BACKED_new_PC_mask;
 
  Running = true;
 
@@ -229,9 +245,11 @@ pscpu_timestamp_t PS_CPU::RunReal(pscpu_timestamp_t timestamp_in)
    {
     BACKED_PC = PC;
     BACKED_new_PC = new_PC;
+    BACKED_new_PC_mask = new_PC_mask;
     CPUHook(PC);
     PC = BACKED_PC;
     new_PC = BACKED_new_PC;
+    new_PC_mask = BACKED_new_PC_mask;
    }
 
 
@@ -286,6 +304,19 @@ meowmoo &= 1;
    #define END_OPF goto OpDone; }
 #endif
 
+   #define DO_BRANCH(offset, mask)			\
+	{						\
+	 PC = (PC & new_PC_mask) + new_PC;		\
+	 new_PC = (offset);				\
+	 new_PC_mask = (mask);				\
+         if(DebugMode && ADDBT)                 	\
+	 {						\
+          ADDBT(PC, (PC & new_PC_mask) + new_PC, false);	\
+	 }						\
+	 InBDSlot = true;				\
+	 goto SkipNPCStuff;				\
+	}
+#if 0
    #define DO_BRANCH(branch_target)		\
 	{					\
 	 uint32 bt_tmp = (branch_target);	\
@@ -304,7 +335,7 @@ meowmoo &= 1;
          InBDSlot = true;			\
          goto SkipNPCStuff;			\
 	}
-
+#endif
 
    #define ITYPE uint32 rs __attribute__((unused)) = (instr >> 21) & 0x1F; uint32 rt __attribute__((unused)) = (instr >> 16) & 0x1F; int32 immediate = (int16)(instr & 0xFFFF); /*printf(" rs=%02x(%08x), rt=%02x(%08x), immediate=(%08x) ", rs, GPR[rs], rt, GPR[rt], immediate);*/
    #define ITYPE_ZE uint32 rs __attribute__((unused)) = (instr >> 21) & 0x1F; uint32 rt __attribute__((unused)) = (instr >> 16) & 0x1F; uint32 immediate = instr & 0xFFFF; /*printf(" rs=%02x(%08x), rt=%02x(%08x), immediate=(%08x) ", rs, GPR[rs], rt, GPR[rt], immediate);*/
@@ -359,6 +390,7 @@ meowmoo &= 1;
 	     PSX_WARNING("[CPU] Unknown instruction @%08x = %08x, op=%02x, funct=%02x", PC, instr, instr >> 26, (instr & 0x3F));
 	     DO_LDS();
 	     new_PC = Exception(EXCEPTION_RI, PC);
+	     new_PC_mask = 0;
     END_OPF;
 
     //
@@ -372,7 +404,10 @@ meowmoo &= 1;
 	DO_LDS();
 
 	if(ep)
+	{
 	 new_PC = Exception(EXCEPTION_OV, PC);
+         new_PC_mask = 0;
+	}
 	else
 	 GPR[rd] = result;
 
@@ -389,7 +424,10 @@ meowmoo &= 1;
 	DO_LDS();
 
         if(ep)
+	{
 	 new_PC = Exception(EXCEPTION_OV, PC);
+         new_PC_mask = 0;
+	}
         else
          GPR[rt] = result;
 
@@ -458,7 +496,7 @@ meowmoo &= 1;
 
 	if(result)
 	{
-	 DO_BRANCH((PC + 4) + (immediate << 2));
+	 DO_BRANCH((immediate << 2), ~0U);
 	}
     END_OPF;
 
@@ -479,7 +517,7 @@ meowmoo &= 1;
 
         if(result)
 	{
-	 DO_BRANCH((PC + 4) + (immediate << 2));
+	 DO_BRANCH((immediate << 2), ~0U);
 	}
 
     END_OPF;
@@ -496,7 +534,7 @@ meowmoo &= 1;
 
 	if(result)
 	{
-	 DO_BRANCH((PC + 4) + (immediate << 2));
+	 DO_BRANCH((immediate << 2), ~0U);
 	}
     END_OPF;
 
@@ -511,7 +549,7 @@ meowmoo &= 1;
 
 	if(result)
 	{
-	 DO_BRANCH((PC + 4) + (immediate << 2));
+	 DO_BRANCH((immediate << 2), ~0U);
 	}
 
     END_OPF;
@@ -527,7 +565,7 @@ meowmoo &= 1;
 
 	if(result)
 	{
-	 DO_BRANCH((PC + 4) + (immediate << 2));
+	 DO_BRANCH((immediate << 2), ~0U);
 	}
 
     END_OPF;
@@ -536,8 +574,11 @@ meowmoo &= 1;
     // BREAK - Breakpoint
     //
     BEGIN_OPF(BREAK, 0, 0x0D);
+	PSX_WARNING("[CPU] BREAK BREAK BREAK BREAK DAAANCE -- PC=0x%08x", PC);
+
 	DO_LDS();
 	new_PC = Exception(EXCEPTION_BP, PC);
+        new_PC_mask = 0;
     END_OPF;
 
     // Cop "instructions":	CFCz(no CP0), COPz, CTCz(no CP0), LWCz(no CP0), MFCz, MTCz, SWCz(no CP0)
@@ -561,15 +602,11 @@ meowmoo &= 1;
 		 uint32 rt = (instr >> 16) & 0x1F;
 		 uint32 rd = (instr >> 11) & 0x1F;
 
-		 assert(rd != 15);
-		 assert(rd != 3);
-		 assert(rd != 9);
-		 assert(rd != 11);
-
 		 //printf("MFC0: rt=%d <- rd=%d(%08x)\n", rt, rd, CP0.Regs[rd]);
 		 DO_LDS();
 
-		 GPR[rt] = CP0.Regs[rd];
+		 LDWhich = rt;
+		 LDValue = CP0.Regs[rd];
 		}
 		break;
 
@@ -579,11 +616,37 @@ meowmoo &= 1;
 		 uint32 rd = (instr >> 11) & 0x1F;
 		 uint32 val = GPR[rt];
 
-		 if(rd != 12)
-	 	  printf("MTC0: rt=%d(%08x) -> rd=%d\n", rt, GPR[rt], rd);
+		 if(rd != CP0REG_CAUSE && rd != CP0REG_SR && val)
+		 {
+	 	  PSX_WARNING("[CPU] Unimplemented MTC0: rt=%d(%08x) -> rd=%d", rt, GPR[rt], rd);
+		 }
 
 		 switch(rd)
 		 {
+		  case CP0REG_BPC:
+			CP0.BPC = val;
+			break;
+
+		  case CP0REG_BDA:
+			CP0.BDA = val;
+			break;
+
+		  case CP0REG_TAR:
+			CP0.TAR = val;
+			break;
+
+		  case CP0REG_DCIC:
+			CP0.DCIC = val & 0xFF80003F;
+			break;
+
+  		  case CP0REG_BDAM:
+			CP0.BDAM = val;
+			break;
+
+  		  case CP0REG_BPCM:
+			CP0.BPCM = val;
+			break;
+
 		  case CP0REG_CAUSE:
 			CP0.CAUSE &= ~(0x3 << 8);
 			CP0.CAUSE |= val & (0x3 << 8);
@@ -617,6 +680,7 @@ meowmoo &= 1;
     BEGIN_OPF(COP1, 0x11, 0);
 	DO_LDS();
         new_PC = Exception(EXCEPTION_COPU, PC);
+        new_PC_mask = 0;
     END_OPF;
 
     //
@@ -641,7 +705,8 @@ meowmoo &= 1;
 
  	 	 DO_LDS();
 
-		 GPR[rt] = GTE_ReadDR(rd);
+		 LDWhich = rt;
+		 LDValue = GTE_ReadDR(rd);
 		}
 		break;
 
@@ -654,6 +719,7 @@ meowmoo &= 1;
 	         if(timestamp < gte_ts_done)
 	          timestamp = gte_ts_done;
 
+		 //printf("GTE WriteDR: %d %d\n", rd, val);
 		 GTE_WriteDR(rd, val);
 	         DO_LDS();
 		}
@@ -669,7 +735,8 @@ meowmoo &= 1;
 
 	 	 DO_LDS();
 
-		 GPR[rt] = GTE_ReadCR(rd);
+		 LDWhich = rt;
+		 LDValue = GTE_ReadCR(rd);
 		//printf("GTE ReadCR: %d %d\n", rd, GPR[rt]);
 		}		
 		break;
@@ -705,6 +772,7 @@ meowmoo &= 1;
     BEGIN_OPF(COP3, 0x13, 0);
 	DO_LDS();
         new_PC = Exception(EXCEPTION_COPU, PC);
+        new_PC_mask = 0;
     END_OPF;
 
     //
@@ -713,6 +781,7 @@ meowmoo &= 1;
     BEGIN_OPF(LWC0, 0x30, 0);
 	DO_LDS();
         new_PC = Exception(EXCEPTION_COPU, PC);
+        new_PC_mask = 0;
     END_OPF;
 
     //
@@ -721,6 +790,7 @@ meowmoo &= 1;
     BEGIN_OPF(LWC1, 0x31, 0);
 	DO_LDS();
         new_PC = Exception(EXCEPTION_COPU, PC);
+        new_PC_mask = 0;
     END_OPF;
 
     //
@@ -731,7 +801,10 @@ meowmoo &= 1;
         uint32 address = GPR[rs] + immediate;
 
         if(address & 3)
+	{
          new_PC = Exception(EXCEPTION_ADEL, PC);
+         new_PC_mask = 0;
+	}
         else
 	{
          if(timestamp < gte_ts_done)
@@ -749,6 +822,7 @@ meowmoo &= 1;
     BEGIN_OPF(LWC3, 0x33, 0);
 	DO_LDS();
         new_PC = Exception(EXCEPTION_COPU, PC);
+        new_PC_mask = 0;
     END_OPF;
 
 
@@ -758,6 +832,7 @@ meowmoo &= 1;
     BEGIN_OPF(SWC0, 0x38, 0);
 	DO_LDS();
 	new_PC = Exception(EXCEPTION_COPU, PC);
+        new_PC_mask = 0;
     END_OPF;
 
     //
@@ -766,6 +841,7 @@ meowmoo &= 1;
     BEGIN_OPF(SWC1, 0x39, 0);
 	DO_LDS();
         new_PC = Exception(EXCEPTION_COPU, PC);
+        new_PC_mask = 0;
     END_OPF;
 
     //
@@ -776,7 +852,10 @@ meowmoo &= 1;
         uint32 address = GPR[rs] + immediate;
 
 	if(address & 0x3)
+	{
 	 new_PC = Exception(EXCEPTION_ADES, PC);
+         new_PC_mask = 0;
+	}
 	else if(!(CP0.SR & 0x10000))
 	{
          if(timestamp < gte_ts_done)
@@ -793,6 +872,7 @@ meowmoo &= 1;
     BEGIN_OPF(SWC3, 0x3B, 0);
 	DO_LDS();
         new_PC = Exception(EXCEPTION_RI, PC);
+        new_PC_mask = 0;
     END_OPF;
 
 
@@ -859,7 +939,7 @@ meowmoo &= 1;
 
 	DO_LDS();
 
-	DO_BRANCH((target << 2) | ((PC + 4) & 0xF0000000));
+	DO_BRANCH(target << 2, 0xF0000000);
     END_OPF;
 
     //
@@ -872,7 +952,7 @@ meowmoo &= 1;
 
 	GPR[31] = PC + 8;
 
-	DO_BRANCH((target << 2) | ((PC + 4) & 0xF0000000));
+	DO_BRANCH(target << 2, 0xF0000000);
     END_OPF;
 
     //
@@ -886,7 +966,7 @@ meowmoo &= 1;
 
 	GPR[rd] = PC + 8;
 
-	DO_BRANCH(tmp);
+	DO_BRANCH(tmp, 0);
 
     END_OPF;
 
@@ -899,7 +979,7 @@ meowmoo &= 1;
 
 	DO_LDS();
 
-	DO_BRANCH(bt);
+	DO_BRANCH(bt, 0);
 
     END_OPF;
 
@@ -1189,7 +1269,10 @@ meowmoo &= 1;
 	DO_LDS();
 
 	if(ep)
+	{
 	 new_PC = Exception(EXCEPTION_OV, PC);
+         new_PC_mask = 0;
+	}
 	else
 	 GPR[rd] = result;
 
@@ -1217,6 +1300,7 @@ meowmoo &= 1;
 	DO_LDS();
 
 	new_PC = Exception(EXCEPTION_SYSCALL, PC);
+        new_PC_mask = 0;
     END_OPF;
 
 
@@ -1285,7 +1369,10 @@ meowmoo &= 1;
 	DO_LDS();
 
 	if(address & 1)
+	{
 	 new_PC = Exception(EXCEPTION_ADEL, PC);
+         new_PC_mask = 0;
+	}
 	else
 	{
 	 LDWhich = rt;
@@ -1303,7 +1390,10 @@ meowmoo &= 1;
 	DO_LDS();
 
         if(address & 1)
+	{
          new_PC = Exception(EXCEPTION_ADEL, PC);
+         new_PC_mask = 0;
+	}
 	else
 	{
 	 LDWhich = rt;
@@ -1322,7 +1412,10 @@ meowmoo &= 1;
 	DO_LDS();
 
         if(address & 3)
+	{
          new_PC = Exception(EXCEPTION_ADEL, PC);
+         new_PC_mask = 0;
+	}
         else
 	{
 	 LDWhich = rt;
@@ -1350,7 +1443,10 @@ meowmoo &= 1;
         uint32 address = GPR[rs] + immediate;
 
 	if(address & 0x1)
+	{
 	 new_PC = Exception(EXCEPTION_ADES, PC);
+         new_PC_mask = 0;
+	}
 	else
 	 WriteMemory<uint16>(timestamp, address, GPR[rt]);
 
@@ -1365,14 +1461,17 @@ meowmoo &= 1;
         uint32 address = GPR[rs] + immediate;
 
 	if(address & 0x3)
+	{
 	 new_PC = Exception(EXCEPTION_ADES, PC);
+         new_PC_mask = 0;
+	}
 	else
 	 WriteMemory<uint32>(timestamp, address, GPR[rt]);
 
 	DO_LDS();
     END_OPF;
 
-    // Be careful with DO_LDS placement with LWL/SWL/LWR/SWR instructions, I think.  It's probably not correct now, too.
+    // LWL and LWR load delay slot tomfoolery appears to apply even to MFC0! (and probably MFCn and CFCn as well, though they weren't explicitly tested)
 
     //
     // LWL - Load Word Left
@@ -1380,25 +1479,30 @@ meowmoo &= 1;
     BEGIN_OPF(LWL, 0x22, 0);
 	ITYPE;
 	uint32 address = GPR[rs] + immediate;
+	uint32 v = GPR[rt];
 
-	DO_LDS();
+	if(LDWhich == rt)
+	{
+	 v = LDValue;
+	}
+	else
+	{
+	 DO_LDS();
+	}
 
+	LDWhich = rt;
 	switch(address & 0x3)
 	{
-	 case 0: GPR[rt] &= ~(0xFF << 24);
-		 GPR[rt] |= ReadMemory<uint8>(timestamp, address & ~3) << 24;
+	 case 0: LDValue = (v & ~(0xFF << 24)) | (ReadMemory<uint8>(timestamp, address & ~3) << 24);
 		 break;
 
-	 case 1: GPR[rt] &= ~(0xFFFF << 16);
-		 GPR[rt] |= ReadMemory<uint16>(timestamp, address & ~3) << 16;
+	 case 1: LDValue = (v & ~(0xFFFF << 16)) | (ReadMemory<uint16>(timestamp, address & ~3) << 16);
 	         break;
 
-	 case 2: GPR[rt] &= ~(0xFFFFFF << 8);
-		 GPR[rt] |= ReadMemory<uint32>(timestamp, address & ~3, true) << 8;
+	 case 2: LDValue = (v & ~(0xFFFFFF << 8)) | (ReadMemory<uint32>(timestamp, address & ~3, true) << 8);
 		 break;
 
-	 case 3: GPR[rt] &= ~(0xFFFFFFFF << 0);
-		 GPR[rt] |= ReadMemory<uint32>(timestamp, address & ~3) << 0;
+	 case 3: LDValue = (v & ~(0xFFFFFFFF << 0)) | (ReadMemory<uint32>(timestamp, address & ~3) << 0);
 		 break;
 	}
     END_OPF;
@@ -1434,25 +1538,30 @@ meowmoo &= 1;
     BEGIN_OPF(LWR, 0x26, 0);
         ITYPE;
         uint32 address = GPR[rs] + immediate;
+	uint32 v = GPR[rt];
 
-	DO_LDS();
+	if(LDWhich == rt)
+	{
+	 v = LDValue;
+	}
+	else
+	{
+	 DO_LDS();
+	}
 
+	LDWhich = rt;
 	switch(address & 0x3)
 	{
-	 case 0: GPR[rt] &= ~(0xFFFFFFFF);
-		 GPR[rt] |= ReadMemory<uint32>(timestamp, address);
+	 case 0: LDValue = (v & ~(0xFFFFFFFF)) | ReadMemory<uint32>(timestamp, address);
 		 break;
 
-	 case 1: GPR[rt] &= ~(0xFFFFFF);
-		 GPR[rt] |= ReadMemory<uint32>(timestamp, address, true);
+	 case 1: LDValue = (v & ~(0xFFFFFF)) | ReadMemory<uint32>(timestamp, address, true);
 		 break;
 
-	 case 2: GPR[rt] &= ~(0xFFFF);
-		 GPR[rt] |= ReadMemory<uint16>(timestamp, address);
+	 case 2: LDValue = (v & ~(0xFFFF)) | ReadMemory<uint16>(timestamp, address);
 	         break;
 
-	 case 3: GPR[rt] &= ~(0xFF);
-		 GPR[rt] |= ReadMemory<uint8>(timestamp, address);
+	 case 3: LDValue = (v & ~(0xFF)) | ReadMemory<uint8>(timestamp, address);
 		 break;
 	}
     END_OPF;
@@ -1487,17 +1596,26 @@ meowmoo &= 1;
     // Mednafen special instruction
     //
     BEGIN_OPF(INTERRUPT, 0x3F, 0);
-	DO_LDS();
+	if(Halted)
+	{
+	 goto SkipNPCStuff;
+	}
+	else
+	{
+ 	 DO_LDS();
 
-	new_PC = Exception(EXCEPTION_INT, PC);
+	 new_PC = Exception(EXCEPTION_INT, PC);
+         new_PC_mask = 0;
+	}
     END_OPF;
    }
 
    OpDone: ;
 
    InBDSlot = false;
-   PC = new_PC;
-   new_PC += 4;
+   PC = (PC & new_PC_mask) + new_PC;
+   new_PC_mask = ~0U;
+   new_PC = 4;
 
    SkipNPCStuff:	;
 
@@ -1512,6 +1630,7 @@ meowmoo &= 1;
 
  BACKED_PC = PC;
  BACKED_new_PC = new_PC;
+ BACKED_new_PC_mask = new_PC_mask;
 
  return(timestamp);
 }
@@ -1577,7 +1696,6 @@ void PS_CPU::SetRegister(unsigned int which, uint32 value)
  {
   case GSREG_PC:
         BACKED_PC = value & ~0x3;	// Remove masking if we ever add proper misaligned PC exception
-	BACKED_new_PC = BACKED_PC + 4;
         break;
 
   case GSREG_LO:
