@@ -51,6 +51,8 @@ class PS_CPU
  // which ranges 0-5, inclusive
  void AssertIRQ(int which, bool asserted);
 
+ void SetHalt(bool status);
+
  private:
 
  struct
@@ -64,12 +66,18 @@ class PS_CPU
 
  uint32 BACKED_PC;
  uint32 BACKED_new_PC;
+ uint32 BACKED_new_PC_mask;
 
  uint32 IPCache;
  void RecalcIPCache(void);
+ bool Halted;
 
  uint32 LDWhich;
  uint32 LDValue;
+ uint32 LDPenalty;
+
+ uint32 ReadPenaltyCounter;
+ uint32 ReadPenaltyMask;
 
  bool InBDSlot;
 
@@ -82,22 +90,16 @@ class PS_CPU
 
  enum
  {
-  CP0REG_INX = 0,
-  CP0REG_RAND = 1,
-  CP0REG_TLBLO = 2,
-  CP0REG_BPC = 3,
-  CP0REG_CTXT = 4,
-  CP0REG_BDA = 5,
-  CP0REG_PIDMASK = 6,
-  CP0REG_DCIC = 7,
-  CP0REG_BADVADDR = 8,
-  CP0REG_BDAM = 9,
-  CP0REG_PID = 10,
-  CP0REG_BPCM = 11,
+  CP0REG_BPC = 3,		// PC breakpoint address.
+  CP0REG_BDA = 5,		// Data load/store breakpoint address.
+  CP0REG_TAR = 6,		// Target address(???)
+  CP0REG_DCIC = 7,		// Cache control
+  CP0REG_BDAM = 9,		// Data load/store address mask.
+  CP0REG_BPCM = 11,		// PC breakpoint address mask.
   CP0REG_SR = 12,
   CP0REG_CAUSE = 13,
   CP0REG_EPC = 14,
-  CP0REG_PRID = 15,
+  CP0REG_PRID = 15,		// Product ID
   CP0REG_ERREG = 16
  };
 
@@ -108,23 +110,23 @@ class PS_CPU
    uint32 Regs[32];
    struct
    {
-    uint32 INX;
-    uint32 RAND;
-    uint32 TLBLO;
+    uint32 Unused00;
+    uint32 Unused01;
+    uint32 Unused02;
     uint32 BPC;		// RW
-    uint32 CTXT;
+    uint32 Unused04;
     uint32 BDA;		// RW
-    uint32 PIDMASK;
-    uint32 DCIC;		// RW
-    uint32 BADVADDR;		// R
-    uint32 BDAM;		// R/W
-    uint32 PID;
-    uint32 BPCMASK;		// R/W
-    uint32 SR;			// R/W
-    uint32 CAUSE;		// R
+    uint32 TAR;
+    uint32 DCIC;	// RW
+    uint32 Unused08;	
+    uint32 BDAM;	// R/W
+    uint32 Unused0A;
+    uint32 BPCM;	// R/W
+    uint32 SR;		// R/W
+    uint32 CAUSE;	// R/W(partial)
     uint32 EPC;		// R
-    uint32 PRID;		// R
-    uint32 ERREG;
+    uint32 PRID;	// R
+    uint32 ERREG;	// ?(may not exist, test)
    };
   };
  } CP0;
