@@ -1,9 +1,30 @@
+/* Mednafen - Multi-system Emulator
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ */
 
 #ifndef PSXDEV_GTE_TESTING
 #include "psx.h"
 #include "gte.h"
 #endif
 
+static uint32 ReciprocalTable[0x8000] =
+{
+ #include "gte_divrecip.inc"
+};
+
 /* Notes:
 
  AVSZ3/AVSZ4:
@@ -49,13 +70,10 @@ typedef struct
  int16 Y;
 } gtexy;
 
-int32 A(unsigned int which, int64 value);
 int16 Lm_B(unsigned int which, int32 value, int lm);
 uint8 Lm_C(unsigned int which, int32 value);
 
-int32 Lm_D(int32 value);
 
-int32 F(int64 value);
 
 int32 Lm_G(unsigned int which, int32 value);
 int32 Lm_H(int32 value);
@@ -77,7 +95,7 @@ typedef union
   gtematrix Rot;
   gtematrix Light;
   gtematrix Color;
-  gtematrix Null;	// Might not be correct.
+  gtematrix AbbyNormal;
  };
 } Matrices_t;
 
@@ -110,23 +128,24 @@ static int16 ZSF4;
 static int16 Vectors[3][4];
 static gtergb RGB;
 static uint16 OTZ;
-static int16 IR0;
-static int16 IR1;
-static int16 IR2;
-static int16 IR3;
+
+static int16 IR[4];
+
+#define IR0 IR[0]
+#define IR1 IR[1]
+#define IR2 IR[2]
+#define IR3 IR[3]
+
 static gtexy XY_FIFO[4];
 static uint16 Z_FIFO[4];
 static gtergb RGB_FIFO[3];
 static int32 MAC[4];
-static uint32 IRGB;
 static uint32 LZCS;
 static uint32 LZCR;
 
 static uint32 Reg23;
 // end DR
 
-void PTransform(uint32 sf, int lm, unsigned int v);
-
 int32 RTPS(uint32 instr);
 int32 RTPT(uint32 instr);
 
@@ -201,7 +220,6 @@ void GTE_Power(void)
  memset(Z_FIFO, 0, sizeof(Z_FIFO));
  memset(RGB_FIFO, 0, sizeof(RGB_FIFO));
  memset(MAC, 0, sizeof(MAC));
- IRGB = 0;
  LZCS = 0;
  LZCR = 0;
 
@@ -397,6 +415,8 @@ void GTE_WriteDR(unsigned int which, uint32 value)
   case 14:
 	XY_FIFO[2].X = value;
 	XY_FIFO[2].Y = value >> 16;
+	XY_FIFO[3].X = value;
+	XY_FIFO[3].Y = value >> 16;
 	break;
 
   case 15:
@@ -466,7 +486,6 @@ void GTE_WriteDR(unsigned int which, uint32 value)
 	break;
 
   case 28:
-	IRGB = value & 0x7FFF;
 	IR1 = ((value >> 0) & 0x1F) << 7;
 	IR2 = ((value >> 5) & 0x1F) << 7;
 	IR3 = ((value >> 10) & 0x1F) << 7;
@@ -476,7 +495,6 @@ void GTE_WriteDR(unsigned int which, uint32 value)
 	break;
 
   case 30:
-	// FIXME
 	LZCS = value;
 	{
 	 uint32 test = value & 0x80000000;
@@ -614,9 +632,6 @@ uint32 GTE_ReadDR(unsigned int which)
 	break;
 
   case 28:
-	ret = IRGB;
-	break;
-
   case 29:
 	ret = Sat5(IR1 >> 7) | (Sat5(IR2 >> 7) << 5) | (Sat5(IR3 >> 7) << 10);
 	break;
@@ -632,30 +647,19 @@ uint32 GTE_ReadDR(unsigned int which)
  return(ret);
 }
 
-INLINE int32 A(unsigned int which, int64 value)
+#define sign_x_to_s64(_bits, _value) (((int64)((uint64)(_value) << (64 - _bits))) >> (64 - _bits))
+INLINE int64 A_MV(unsigned which, int64 value)
 {
- // Done for an issue with NCCS, at least.  See if this masking-out is applicable in all cases, and for other registers.
- //
- FLAGS &= ~(1 << (27 - which));
- FLAGS &= ~(1 << (30 - which));
- //
- //
-
- if(value < -2147483648LL)
- {
-  // flag set here
-  FLAGS |= 1 << (27 - which);
- }
-
- if(value > 2147483647LL)
- {
-  // flag set here
+ if(value >= (1LL << 43))
   FLAGS |= 1 << (30 - which);
- }
- return(value);
+
+ if(value < -(1LL << 43))
+  FLAGS |= 1 << (27 - which);
+
+ return sign_x_to_s64(44, value);
 }
 
-INLINE int32 F(int64 value)
+INLINE int64 F(int64 value)
 {
  if(value < -2147483648LL)
  {
@@ -693,6 +697,34 @@ INLINE int16 Lm_B(unsigned int which, int32 value, int lm)
  return(value);
 }
 
+
+INLINE int16 Lm_B_PTZ(unsigned int which, int32 value, int32 ftv_value, int lm)
+{
+ int32 tmp = lm << 15;
+
+ if(ftv_value < -32768)
+ {
+  FLAGS |= 1 << (24 - which);
+ }
+
+ if(ftv_value > 32767)
+ {
+  FLAGS |= 1 << (24 - which);
+ }
+
+ if(value < (-32768 + tmp))
+ {
+  value = -32768 + tmp;
+ }
+
+ if(value > 32767)
+ {
+  value = 32767;
+ }
+
+ return(value);
+}
+
 INLINE uint8 Lm_C(unsigned int which, int32 value)
 {
  if(value & ~0xFF)
@@ -710,19 +742,22 @@ INLINE uint8 Lm_C(unsigned int which, int32 value)
  return(value);
 }
 
-INLINE int32 Lm_D(int32 value)
+INLINE int32 Lm_D(int32 value, int unchained)
 {
  // Not sure if we should have it as int64, or just chain on to and special case when the F flags are set.
- if(FLAGS & (1 << 15))
+ if(!unchained)
  {
-  FLAGS |= 1 << 18;
-  return(0);
- }
+  if(FLAGS & (1 << 15))
+  {
+   FLAGS |= 1 << 18;
+   return(0);
+  }
 
- if(FLAGS & (1 << 16))
- {
-  FLAGS |= 1 << 18;
-  return(0xFFFF);
+  if(FLAGS & (1 << 16))
+  {
+   FLAGS |= 1 << 18;
+   return(0xFFFF);
+  }
  }
 
  if(value < 0)
@@ -760,18 +795,33 @@ INLINE int32 Lm_G(unsigned int which, int32 value)
  return(value);
 }
 
+// limit to 4096, not 4095
 INLINE int32 Lm_H(int32 value)
 {
+#if 0
+ if(FLAGS & (1 << 15))
+ {
+  value = 0;
+  FLAGS |= 1 << 12;
+  return value;
+ }
+
+ if(FLAGS & (1 << 16))
+ {
+  value = 4096;
+  FLAGS |= 1 << 12;
+  return value;
+ }
+#endif
+
  if(value < 0)
  {
-  // Set flag here
   value = 0;
   FLAGS |= 1 << 12;
  }
 
- if(value > 4096)	// Yes, six.
+ if(value > 4096)
  {
-  // Set flag here
   value = 4096;
   FLAGS |= 1 << 12;
  }
@@ -797,38 +847,91 @@ INLINE void MAC_to_IR(int lm)
  IR3 = Lm_B(2, MAC[3], lm);
 }
 
-// NOTES:
-// << 12 for translation vector with MVMA when sf is 0.
-// FIXME: far color vector is borked
 INLINE void MultiplyMatrixByVector(const gtematrix *matrix, const int16 *v, const int32 *crv, uint32 sf, int lm)
 {
-#if 0
- MAC[1] = A(0, ((int64)crv[0] << 12) + (((int64)(matrix.MX[0][0] * v[0]) + (matrix.MX[0][1] * v[1]) + (matrix.MX[0][2] * v[2])) >> sf));
- MAC[2] = A(1, ((int64)crv[1] << 12) + (((int64)(matrix.MX[1][0] * v[0]) + (matrix.MX[1][1] * v[1]) + (matrix.MX[1][2] * v[2])) >> sf));
- MAC[3] = A(2, ((int64)crv[2] << 12) + (((int64)(matrix.MX[2][0] * v[0]) + (matrix.MX[2][1] * v[1]) + (matrix.MX[2][2] * v[2])) >> sf));
-#endif
+ unsigned i;
 
- assert(crv != CRVectors.FC);
+ for(i = 0; i < 3; i++)
+ {
+  int64 tmp;
+  int32 mulr[3];
 
-#if 0
- if(0 && crv == CRVectors.FC)
- {
-  MAC[1] = A(0, ((((int64)crv[0] << 8) + (int64)(matrix->MX[0][0] * v[0]) + (matrix->MX[0][1] * v[1]) + (matrix->MX[0][2] * v[2])) >> sf));
-  MAC[2] = A(1, ((((int64)crv[1] << 8) + (int64)(matrix->MX[1][0] * v[0]) + (matrix->MX[1][1] * v[1]) + (matrix->MX[1][2] * v[2])) >> sf));
-  MAC[3] = A(2, ((((int64)crv[2] << 8) + (int64)(matrix->MX[2][0] * v[0]) + (matrix->MX[2][1] * v[1]) + (matrix->MX[2][2] * v[2])) >> sf));
- }
- else
-#endif
- {
-  MAC[1] = A(0, ((((int64)crv[0] << 12) + (int64)(matrix->MX[0][0] * v[0]) + (matrix->MX[0][1] * v[1]) + (matrix->MX[0][2] * v[2])) >> sf));
-  MAC[2] = A(1, ((((int64)crv[1] << 12) + (int64)(matrix->MX[1][0] * v[0]) + (matrix->MX[1][1] * v[1]) + (matrix->MX[1][2] * v[2])) >> sf));
-  MAC[3] = A(2, ((((int64)crv[2] << 12) + (int64)(matrix->MX[2][0] * v[0]) + (matrix->MX[2][1] * v[1]) + (matrix->MX[2][2] * v[2])) >> sf));
+  tmp = (int64)crv[i] << 12;
+
+  if(matrix == &Matrices.AbbyNormal)
+  {
+   if(i == 0)
+   {
+    mulr[0] = -((RGB.R << 4) * v[0]);
+    mulr[1] = (RGB.R << 4) * v[1];
+    mulr[2] = IR0 * v[2];
+   }
+   else
+   {
+    mulr[0] = (int16)CR[i] * v[0];
+    mulr[1] = (int16)CR[i] * v[1];
+    mulr[2] = (int16)CR[i] * v[2];
+   }
+  }
+  else
+  {
+   mulr[0] = matrix->MX[i][0] * v[0];
+   mulr[1] = matrix->MX[i][1] * v[1];
+   mulr[2] = matrix->MX[i][2] * v[2];
+  }
+
+  tmp = A_MV(i, tmp + mulr[0]);
+  if(crv == CRVectors.FC)
+  {
+   Lm_B(i, tmp >> sf, FALSE);
+   tmp = 0;
+  }
+
+  tmp = A_MV(i, tmp + mulr[1]);
+  tmp = A_MV(i, tmp + mulr[2]);
+
+  MAC[1 + i] = tmp >> sf;
  }
 
+
  MAC_to_IR(lm);
 }
 
 
+INLINE void MultiplyMatrixByVector_PT(const gtematrix *matrix, const int16 *v, const int32 *crv, uint32 sf, int lm)
+{
+ int64 tmp[3];
+ unsigned i;
+
+ for(i = 0; i < 3; i++)
+ {
+  int32 mulr[3];
+
+  tmp[i] = (int64)crv[i] << 12;
+
+  mulr[0] = matrix->MX[i][0] * v[0];
+  mulr[1] = matrix->MX[i][1] * v[1];
+  mulr[2] = matrix->MX[i][2] * v[2];
+
+  tmp[i] = A_MV(i, tmp[i] + mulr[0]);
+  tmp[i] = A_MV(i, tmp[i] + mulr[1]);
+  tmp[i] = A_MV(i, tmp[i] + mulr[2]);
+
+  MAC[1 + i] = tmp[i] >> sf;
+ }
+
+ IR1 = Lm_B(0, MAC[1], lm);
+ IR2 = Lm_B(1, MAC[2], lm);
+ //printf("FTV: %08x %08x\n", crv[2], (uint32)(tmp[2] >> 12));
+ IR3 = Lm_B_PTZ(2, MAC[3], tmp[2] >> 12, lm);
+
+ Z_FIFO[0] = Z_FIFO[1];
+ Z_FIFO[1] = Z_FIFO[2];
+ Z_FIFO[2] = Z_FIFO[3];
+ Z_FIFO[3] = Lm_D(tmp[2] >> 12, TRUE);
+}
+
+
 #define VAR_UNUSED __attribute__((unused))
 
 #define DECODE_FIELDS							\
@@ -856,10 +959,9 @@ int32 SQR(uint32 instr)
 {
  DECODE_FIELDS;
 
- // Typecast to int64 shouldn't be necessary here (might be able to remove call to A()...
- MAC[1] = A(0, ((IR1 * IR1) >> sf));
- MAC[2] = A(1, ((IR2 * IR2) >> sf));
- MAC[3] = A(2, ((IR3 * IR3) >> sf));
+ MAC[1] = ((IR1 * IR1) >> sf);
+ MAC[2] = ((IR2 * IR2) >> sf);
+ MAC[3] = ((IR3 * IR3) >> sf);
 
  MAC_to_IR(lm);
 
@@ -876,51 +978,67 @@ int32 MVMVA(uint32 instr)
  return(8);
 }
 
-
-INLINE void PTransform(uint32 sf, int lm, unsigned int v)
+static INLINE unsigned CountLeadingZeroU16(uint16 val)
 {
- int64 h_div_sz;
+ unsigned ret = 0;
 
- MultiplyMatrixByVector(&Matrices.Rot, Vectors[v], CRVectors.T, sf, lm);
-
- Z_FIFO[0] = Z_FIFO[1];
- Z_FIFO[1] = Z_FIFO[2];
- Z_FIFO[2] = Z_FIFO[3];
- Z_FIFO[3] = Lm_D(MAC[3]);
-
- // FIXME: division
- if(H < (Z_FIFO[3] * 2))
-  h_div_sz = (((int64)H << 16)) / Z_FIFO[3];
- else
+ while(!(val & 0x8000) && ret < 16)
  {
-  h_div_sz = 0x1FFFF;
-  FLAGS |= 1 << 17;
+  val <<= 1;
+  ret++;
  }
 
- MAC[0] = F(((int64)OFX + IR1 * h_div_sz + 32768) >> 16);
+ return ret;
+}
+
+static INLINE uint32 Divide(uint32 dividend, uint32 divisor)
+{
+ //if((Z_FIFO[3] * 2) > H)
+ if((divisor * 2) > dividend)
+ {
+  unsigned shift_bias = CountLeadingZeroU16(divisor);
+
+  dividend <<= shift_bias;
+  divisor <<= shift_bias;
+
+  return ((int64)dividend * ReciprocalTable[divisor & 0x7FFF] + 32768) >> 16;
+ }
+ else
+ {
+  FLAGS |= 1 << 17;
+  return 0x1FFFF;
+ }
+}
+
+static INLINE void TransformXY(int64 h_div_sz)
+{
+ MAC[0] = F((int64)OFX + IR1 * h_div_sz) >> 16;
  XY_FIFO[3].X = Lm_G(0, MAC[0]);
 
- MAC[0] = F(((int64)OFY + IR2 * h_div_sz + 32768) >> 16);
+ MAC[0] = F((int64)OFY + IR2 * h_div_sz) >> 16;
  XY_FIFO[3].Y = Lm_G(1, MAC[0]);
 
  XY_FIFO[0] = XY_FIFO[1];
  XY_FIFO[1] = XY_FIFO[2];
  XY_FIFO[2] = XY_FIFO[3];
+}
 
-// MAC[0] = F((DQB + ((DQA * h_div_sz + 32768) >> 16)) >> 8 );
-// IR0 = Lm_H(MAC[0]);
-
-// printf("MOO: %d %d %16lld\n", DQB, DQA, (long long)h_div_sz);
-
+static INLINE void TransformDQ(int64 h_div_sz)
+{
  MAC[0] = F((int64)DQB + DQA * h_div_sz);
- IR0 = Lm_H(MAC[0] >> sf);
+ IR0 = Lm_H(((int64)DQB + DQA * h_div_sz) >> 12);
 }
 
 int32 RTPS(uint32 instr)
 {
  DECODE_FIELDS;
+ int64 h_div_sz;
 
- PTransform(sf, lm, 0);
+ MultiplyMatrixByVector_PT(&Matrices.Rot, Vectors[0], CRVectors.T, sf, lm);
+ h_div_sz = Divide(H, Z_FIFO[3]);
+
+ TransformXY(h_div_sz);
+ TransformDQ(h_div_sz);
 
  return(15);
 }
@@ -931,7 +1049,17 @@ int32 RTPT(uint32 instr)
  int i;
 
  for(i = 0; i < 3; i++)
-  PTransform(sf, lm, i);
+ {
+  int64 h_div_sz;
+
+  MultiplyMatrixByVector_PT(&Matrices.Rot, Vectors[i], CRVectors.T, sf, lm);
+  h_div_sz = Divide(H, Z_FIFO[3]);
+
+  TransformXY(h_div_sz);
+
+  if(i == 2)
+   TransformDQ(h_div_sz);
+ }
 
  return(23);
 }
@@ -977,9 +1105,9 @@ INLINE void NormColorColor(uint32 v, uint32 sf, int lm)
  tmp_vector[0] = IR1; tmp_vector[1] = IR2; tmp_vector[2] = IR3;
  MultiplyMatrixByVector(&Matrices.Color, tmp_vector, CRVectors.B, sf, lm);
 
- MAC[1] = A(0, ((RGB.R << 4) * IR1) >> sf);
- MAC[2] = A(1, ((RGB.G << 4) * IR2) >> sf);
- MAC[3] = A(2, ((RGB.B << 4) * IR3) >> sf);
+ MAC[1] = ((RGB.R << 4) * IR1) >> sf;
+ MAC[2] = ((RGB.G << 4) * IR2) >> sf;
+ MAC[3] = ((RGB.B << 4) * IR3) >> sf;
 
  MAC_to_IR(lm);
 
@@ -1008,56 +1136,40 @@ int32 NCCT(uint32 instr)
 
 INLINE void DepthCue(int mult_IR123, int RGB_from_FIFO, uint32 sf, int lm)
 {
- int32 R_temp, G_temp, B_temp;
+ int32 RGB_temp[3];
+ int32 IR_temp[3] = { IR1, IR2, IR3 };
+ int i;
 
  //assert(sf);
 
  if(RGB_from_FIFO)
  {
-  R_temp = RGB_FIFO[0].R << 4;
-  G_temp = RGB_FIFO[0].G << 4;
-  B_temp = RGB_FIFO[0].B << 4;
+  RGB_temp[0] = RGB_FIFO[0].R << 4;
+  RGB_temp[1] = RGB_FIFO[0].G << 4;
+  RGB_temp[2] = RGB_FIFO[0].B << 4;
  }
  else
  {
-  R_temp = RGB.R << 4;
-  G_temp = RGB.G << 4;
-  B_temp = RGB.B << 4;
+  RGB_temp[0] = RGB.R << 4;
+  RGB_temp[1] = RGB.G << 4;
+  RGB_temp[2] = RGB.B << 4;
  }
 
  if(mult_IR123)
  {
-  // Note: Do not put A() here!  We might just want to change this to local temporaries.
-  MAC[1] = (((int64)CRVectors.FC[0] << 12) - R_temp * IR1) >> sf;
-  MAC[2] = (((int64)CRVectors.FC[1] << 12) - G_temp * IR2) >> sf;
-  MAC[3] = (((int64)CRVectors.FC[2] << 12) - B_temp * IR3) >> sf;
-
-  MAC[1] = A(0, (R_temp * IR1 + IR0 * Lm_B(0, MAC[1], FALSE)) >> sf);
-  MAC[2] = A(1, (G_temp * IR2 + IR0 * Lm_B(1, MAC[2], FALSE)) >> sf);
-  MAC[3] = A(2, (B_temp * IR3 + IR0 * Lm_B(2, MAC[3], FALSE)) >> sf);
-
-/*
-  MAC[1] = A(0, (int64)((R_temp * IR1) >> sf) + IR0 * Lm_B(0, (int64)CRVectors.FC[0] - ((R_temp * IR1) >> sf), FALSE) );
-  MAC[2] = A(1, (int64)((G_temp * IR2) >> sf) + IR0 * Lm_B(1, (int64)CRVectors.FC[1] - ((G_temp * IR2) >> sf), FALSE) );
-  MAC[3] = A(2, (int64)((B_temp * IR3) >> sf) + IR0 * Lm_B(2, (int64)CRVectors.FC[2] - ((B_temp * IR3) >> sf), FALSE) );
-*/
+  for(i = 0; i < 3; i++)
+  {
+   MAC[1 + i] = A_MV(i, (((int64)CRVectors.FC[i] << 12) - RGB_temp[i] * IR_temp[i])) >> sf;
+   MAC[1 + i] = A_MV(i, (RGB_temp[i] * IR_temp[i] + IR0 * Lm_B(i, MAC[1 + i], FALSE))) >> sf;
+  }
  }
  else
  {
-  // Note: Do not put A() here!  We might just want to change this to local temporaries.
-  MAC[1] = (((int64)CRVectors.FC[0] << 12) - (R_temp << 12)) >> sf;
-  MAC[2] = (((int64)CRVectors.FC[1] << 12) - (G_temp << 12)) >> sf;
-  MAC[3] = (((int64)CRVectors.FC[2] << 12) - (B_temp << 12)) >> sf;
-
-  MAC[1] = A(0, (((int64)R_temp << 12) + IR0 * Lm_B(0, MAC[1], FALSE)) >> sf);
-  MAC[2] = A(1, (((int64)G_temp << 12) + IR0 * Lm_B(1, MAC[2], FALSE)) >> sf);
-  MAC[3] = A(2, (((int64)B_temp << 12) + IR0 * Lm_B(2, MAC[3], FALSE)) >> sf);
-
-/*
-  MAC[1] = A(0, (int64)R_temp + ((IR0 * Lm_B(0, (int64)CRVectors.FC[0] - R_temp, FALSE)) >> sf) );
-  MAC[2] = A(1, (int64)G_temp + ((IR0 * Lm_B(1, (int64)CRVectors.FC[1] - G_temp, FALSE)) >> sf) );
-  MAC[3] = A(2, (int64)B_temp + ((IR0 * Lm_B(2, (int64)CRVectors.FC[2] - B_temp, FALSE)) >> sf) );
-*/
+  for(i = 0; i < 3; i++)
+  {
+   MAC[1 + i] = A_MV(i, (((int64)CRVectors.FC[i] << 12) - (RGB_temp[i] << 12))) >> sf;
+   MAC[1 + i] = A_MV(i, (((int64)RGB_temp[i] << 12) + IR0 * Lm_B(i, MAC[1 + i], FALSE))) >> sf;
+  }
  }
 
  MAC_to_IR(lm);
@@ -1098,29 +1210,17 @@ int32 DPCT(uint32 instr)
  return(17);
 }
 
-// SF field *is* used(tested), but it's a bit...weird.
 int32 INTPL(uint32 instr)
 {
  DECODE_FIELDS;
 
- //if(sf)
- //{
- // MAC[1] = A(0, (int64)IR1 + ((IR0 * Lm_B(0, (int64)CRVectors.FC[0] - IR1, FALSE)) >> 12) );
- // MAC[2] = A(1, (int64)IR2 + ((IR0 * Lm_B(1, (int64)CRVectors.FC[1] - IR2, FALSE)) >> 12) );
- // MAC[3] = A(2, (int64)IR3 + ((IR0 * Lm_B(2, (int64)CRVectors.FC[2] - IR3, FALSE)) >> 12) );
- //}
- //else
- //{
+ MAC[1] = A_MV(0, (((int64)CRVectors.FC[0] << 12) - (IR1 << 12))) >> sf;
+ MAC[2] = A_MV(1, (((int64)CRVectors.FC[1] << 12) - (IR2 << 12))) >> sf;
+ MAC[3] = A_MV(2, (((int64)CRVectors.FC[2] << 12) - (IR3 << 12))) >> sf;
 
- // Note: Do not put A() here!  We might just want to change this to local temporaries.
- MAC[1] = (((int64)CRVectors.FC[0] << 12) - (IR1 << 12)) >> sf;
- MAC[2] = (((int64)CRVectors.FC[1] << 12) - (IR2 << 12)) >> sf;
- MAC[3] = (((int64)CRVectors.FC[2] << 12) - (IR3 << 12)) >> sf;
-
- MAC[1] = A(0, (((int64)IR1 << 12) + IR0 * Lm_B(0, MAC[1], FALSE)) >> sf);
- MAC[2] = A(1, (((int64)IR2 << 12) + IR0 * Lm_B(1, MAC[2], FALSE)) >> sf);
- MAC[3] = A(2, (((int64)IR3 << 12) + IR0 * Lm_B(2, MAC[3], FALSE)) >> sf);
-// }
+ MAC[1] = A_MV(0, (((int64)IR1 << 12) + IR0 * Lm_B(0, MAC[1], FALSE)) >> sf);
+ MAC[2] = A_MV(1, (((int64)IR2 << 12) + IR0 * Lm_B(1, MAC[2], FALSE)) >> sf);
+ MAC[3] = A_MV(2, (((int64)IR3 << 12) + IR0 * Lm_B(2, MAC[3], FALSE)) >> sf);
 
  MAC_to_IR(lm);
 
@@ -1157,11 +1257,45 @@ int32 NCDT(uint32 instr)
  DECODE_FIELDS;
 
  for(i = 0; i < 3; i++)
+ {
   NormColorDepthCue(i, sf, lm);
+ }
 
  return(44);
 }
 
+int32 CC(uint32 instr)
+{
+ DECODE_FIELDS;
+ int16 tmp_vector[3];
+
+ tmp_vector[0] = IR1; tmp_vector[1] = IR2; tmp_vector[2] = IR3;
+ MultiplyMatrixByVector(&Matrices.Color, tmp_vector, CRVectors.B, sf, lm);
+
+ MAC[1] = ((RGB.R << 4) * IR1) >> sf;
+ MAC[2] = ((RGB.G << 4) * IR2) >> sf;
+ MAC[3] = ((RGB.B << 4) * IR3) >> sf;
+
+ MAC_to_IR(lm);
+
+ MAC_to_RGB_FIFO();
+
+ return(11);
+}
+
+int32 CDP(uint32 instr)
+{
+ DECODE_FIELDS;
+ int16 tmp_vector[3];
+
+ tmp_vector[0] = IR1; tmp_vector[1] = IR2; tmp_vector[2] = IR3;
+ MultiplyMatrixByVector(&Matrices.Color, tmp_vector, CRVectors.B, sf, lm);
+
+ DepthCue(TRUE, FALSE, sf, lm);
+
+ return(13);
+}
+
 int32 NCLIP(uint32 instr)
 {
  DECODE_FIELDS;
@@ -1172,14 +1306,13 @@ int32 NCLIP(uint32 instr)
  return(8);
 }
 
-// tested, SF field doesn't matter?
 int32 AVSZ3(uint32 instr)
 {
  DECODE_FIELDS;
 
  MAC[0] = F(((int64)ZSF3 * (Z_FIFO[1] + Z_FIFO[2] + Z_FIFO[3])));
 
- OTZ = Lm_D(MAC[0] >> 12);
+ OTZ = Lm_D(MAC[0] >> 12, FALSE);
 
  return(5);
 }
@@ -1190,19 +1323,21 @@ int32 AVSZ4(uint32 instr)
 
  MAC[0] = F(((int64)ZSF4 * (Z_FIFO[0] + Z_FIFO[1] + Z_FIFO[2] + Z_FIFO[3])));
 
- OTZ = Lm_D(MAC[0] >> 12);
+ OTZ = Lm_D(MAC[0] >> 12, FALSE);
 
  return(5);
 }
 
 
+// -32768 * -32768 - 32767 * -32768 = 2147450880
+// (2 ^ 31) - 1 =		      2147483647
 int32 OP(uint32 instr)
 {
  DECODE_FIELDS;
 
- MAC[1] = A(0, ((int64)(Matrices.Rot.MX[1][1] * IR3) - (Matrices.Rot.MX[2][2] * IR2)) >> sf);
- MAC[2] = A(1, ((int64)(Matrices.Rot.MX[2][2] * IR1) - (Matrices.Rot.MX[0][0] * IR3)) >> sf);
- MAC[3] = A(2, ((int64)(Matrices.Rot.MX[0][0] * IR2) - (Matrices.Rot.MX[1][1] * IR1)) >> sf);
+ MAC[1] = ((Matrices.Rot.MX[1][1] * IR3) - (Matrices.Rot.MX[2][2] * IR2)) >> sf;
+ MAC[2] = ((Matrices.Rot.MX[2][2] * IR1) - (Matrices.Rot.MX[0][0] * IR3)) >> sf;
+ MAC[3] = ((Matrices.Rot.MX[0][0] * IR2) - (Matrices.Rot.MX[1][1] * IR1)) >> sf;
 
  MAC_to_IR(lm);
 
@@ -1213,9 +1348,9 @@ int32 GPF(uint32 instr)
 {
  DECODE_FIELDS;
 
- MAC[1] = A(0, ((IR0 * IR1) >> sf));
- MAC[2] = A(1, ((IR0 * IR2) >> sf));
- MAC[3] = A(2, ((IR0 * IR3) >> sf));
+ MAC[1] = (IR0 * IR1) >> sf;
+ MAC[2] = (IR0 * IR2) >> sf;
+ MAC[3] = (IR0 * IR3) >> sf;
 
  MAC_to_IR(lm);
 
@@ -1228,10 +1363,9 @@ int32 GPL(uint32 instr)
 {
  DECODE_FIELDS;
 
-
- MAC[1] = A(0, ((int64)MAC[1] + ((IR0 * IR1) >> sf)));
- MAC[2] = A(1, ((int64)MAC[2] + ((IR0 * IR2) >> sf)));
- MAC[3] = A(2, ((int64)MAC[3] + ((IR0 * IR3) >> sf)));
+ MAC[1] = A_MV(0, ((int64)MAC[1] << sf) + (IR0 * IR1)) >> sf;
+ MAC[2] = A_MV(1, ((int64)MAC[2] << sf) + (IR0 * IR2)) >> sf;
+ MAC[3] = A_MV(2, ((int64)MAC[3] << sf) + (IR0 * IR3)) >> sf;
 
  MAC_to_IR(lm);
 
@@ -1240,13 +1374,14 @@ int32 GPL(uint32 instr)
  return(5);
 }
 
-
 /*
- 
-
-24 23 22 21 20|19|18 17|16 15|14 13|12 11|10| 9  8| 7  6| 5  4  3  2  1  0
-              |sf| mx  |  v  |  cv |-----|lm|-----------|
 
+---------------------------------------------------------------------------------------------
+| 24 23 22 21 20 | 19 | 18 17 | 16 15 | 14 13 | 12  11 | 10 | 9  8  7  6 | 5  4  3  2  1  0 |
+|-------------------------------------------------------------------------------------------|
+|    (unused)    | sf |  mx   |   v   |   cv  |(unused)| lm |  (unused)  |     opcode       |
+---------------------------------------------------------------------------------------------
+ (unused) = unused, ignored
 
  sf = shift 12
 
@@ -1256,103 +1391,265 @@ int32 GPL(uint32 instr)
 
  cv = add vector(translation/back/far color(bugged)/none)
 
+ (unused) = unused, ignored
+
  lm = limit negative results to 0
 
+ (unused) = unused, ignored
+
+ opcode = operation code 
 */
 
 int32 GTE_Instruction(uint32 instr)
 {
+ const unsigned code = instr & 0x3F;
  int32 ret = 1;
 
- //PSX_WARNING("[GTE] Instruction 0x%08x", instr);
-
  FLAGS = 0;
 
- switch(instr & ((0x1F << 20) | 0x3F))
+ switch(code)
  {
-  default: //PSX_WARNING("[GTE] Unknown instruction: 0x%08x, 0x%08x", instr & ~(0x7F << 25), instr & ((0x1F << 20) | 0x3F));
+  default: 
+#ifndef PSXDEV_GTE_TESTING
+	   PSX_WARNING("[GTE] Unknown instruction code: 0x%02x", code);
+#endif
 	   break;
 
-  case 0x0100001:
+  case 0x00:	// alternate?
+  case 0x01:
 	ret = RTPS(instr);
 	break;
 
-  case 0x0200030:
-	ret = RTPT(instr);
+/*
+  case 0x02:	// UNSTABLE?
 	break;
 
-  case 0x0400012:
-	ret = MVMVA(instr);
+  case 0x03:	// UNSTABLE?
 	break;
 
-  case 0x0600029:
-	ret = DCPL(instr);
+  case 0x04:	// Probably simple with v,cv,sf,mx,lm ignored.  Same calculation as 0x3B?
 	break;
 
-  case 0x0700010:	// RR
-	ret = DPCS(instr);
+  case 0x05:	// UNSTABLE?
 	break;
+*/
 
-  case 0x0900011:	// RR
-	ret = INTPL(instr);
-	break;
-
-  case 0x0A00028:
-	ret = SQR(instr);
-	break;
-
-  case 0x0c0001e:
-	ret = NCS(instr);
-	break;
-
-  case 0x0d00020:
-	ret = NCT(instr);
-	break;
-
-  case 0x0e00013:
-	ret = NCDS(instr);
-	break;
-
-  case 0x0f00016:
-	ret = NCDT(instr);
-	break;
-
-  case 0x0F0002A:
-	ret = DPCT(instr);
-	break;
-
-  case 0x100001b:
-	ret = NCCS(instr);
-	break;
-
-  case 0x110003f:
-	ret = NCCT(instr);
-	break;
-
-  case 0x1400006:
+  case 0x06:
 	ret = NCLIP(instr);
 	break;
 
-  case 0x150002d:
-	ret = AVSZ3(instr);
+/*
+  case 0x07:	// UNSTABLE?
 	break;
 
-  case 0x160002E:
-	ret = AVSZ4(instr);
+  case 0x08:	// UNSTABLE?
 	break;
 
-  case 0x170000C:
+  case 0x09:	// UNSTABLE?
+	break;
+
+  case 0x0A:	// UNSTABLE?
+	break;
+
+  case 0x0B:	// UNSTABLE?
+	break;
+
+*/
+
+  case 0x0C:
 	ret = OP(instr);
 	break;
 
-  case 0x190003D:
+/*
+  case 0x0D:	// UNSTABLE?
+	break;
+
+  case 0x0E:	// UNSTABLE?
+	break;
+
+  case 0x0F:	// UNSTABLE?
+	break;
+*/
+
+  case 0x10:
+	ret = DPCS(instr);
+	break;
+
+  case 0x11:
+	ret = INTPL(instr);
+	break;
+
+  case 0x12:
+	ret = MVMVA(instr);
+	break;
+
+  case 0x13:
+	ret = NCDS(instr);
+	break;
+
+  case 0x14:
+	ret = CDP(instr);
+	break;
+
+
+/*
+  case 0x15:	// does one push on RGB FIFO, what else...
+	break;
+*/
+
+  case 0x16:
+	ret = NCDT(instr);
+	break;
+
+/*
+  case 0x17:	// PARTIALLY UNSTABLE(depending on sf or v or cv or mx or lm???), similar behavior under some conditions to 0x16?
+	break;
+
+  case 0x18:
+	break;
+
+  case 0x19:
+	break;
+*/
+
+  case 0x1A:	// Alternate for 0x29?
+	ret = DCPL(instr);
+	break;
+
+  case 0x1B:
+	ret = NCCS(instr);
+	break;
+
+  case 0x1C:
+	ret = CC(instr);
+	break;
+
+/*
+  case 0x1D:
+	break;
+*/
+
+  case 0x1E:
+	ret = NCS(instr);
+	break;
+
+/*
+  case 0x1F:
+	break;
+*/
+
+  case 0x20:
+	ret = NCT(instr);
+	break;
+/*
+  case 0x21:
+	break;
+
+  case 0x22:	// UNSTABLE?
+	break;
+
+  case 0x23:
+	break;
+
+  case 0x24:
+	break;
+
+  case 0x25:
+	break;
+
+  case 0x26:
+	break;
+
+  case 0x27:
+	break;
+*/
+
+  case 0x28:
+	ret = SQR(instr);
+	break;
+
+  case 0x29:
+	ret = DCPL(instr);
+	break;
+
+  case 0x2A:
+	ret = DPCT(instr);
+	break;
+
+/*
+  case 0x2B:
+	break;
+
+  case 0x2C:
+	break;
+*/
+
+  case 0x2D:
+	ret = AVSZ3(instr);
+	break;
+
+  case 0x2E:
+	ret = AVSZ4(instr);
+	break;
+
+/*
+  case 0x2F:	// UNSTABLE?
+	break;
+*/
+
+  case 0x30:
+	ret = RTPT(instr);
+	break;
+
+/*
+  case 0x31:	// UNSTABLE?
+	break;
+
+  case 0x32:	// UNSTABLE?
+	break;
+
+  case 0x33:	// UNSTABLE?
+	break;
+
+  case 0x34:	// UNSTABLE?
+	break;
+
+  case 0x35:	// UNSTABLE?
+	break;
+
+  case 0x36:	// UNSTABLE?
+	break;
+
+  case 0x37:	// UNSTABLE?
+	break;
+
+  case 0x38:
+	break;
+
+  case 0x39:	// Probably simple with v,cv,sf,mx,lm ignored.
+	break;
+
+  case 0x3A:	// Probably simple with v,cv,sf,mx,lm ignored.
+	break;
+
+  case 0x3B:	// Probably simple with v,cv,sf,mx,lm ignored.  Same calculation as 0x04?
+	break;
+
+  case 0x3C:	// UNSTABLE?
+	break;
+*/
+
+  case 0x3D:
 	ret = GPF(instr);
 	break;
 
-  case 0x1A0003E:
+  case 0x3E:
 	ret = GPL(instr);
 	break;
 	
+  case 0x3F:
+	ret = NCCT(instr);
+	break;
  }
 
  if(FLAGS & 0x7f87e000)
@@ -1360,12 +1657,6 @@ int32 GTE_Instruction(uint32 instr)
 
  CR[31] = FLAGS;
 
- //if(timestamp < ts_done)
- // ret = ts_done - timestamp;
-
- // Execute instruction here, and set ts_done
- // ts_done = timestamp + InstrMap[instr]();
-
  return(ret - 1);
 }
 
