@@ -34,9 +34,6 @@
 
 	For ADSR and volume sweep, should the divider be reset to 0 on &0x8000 == true, or should the upper bit be cleared?
 
-	Test exactly what the upper bit of the SPU control register controls.  Even when it == 0, on the real thing, the global sweep/volume updates at least
-	to some degree(haven't fully tested it, assuming the same as when it == 1).  Perhaps it only disables fetching samples from SPURAM or somesuch?
-
 	Should shift occur on all stages of ADPCM sample decoding, or only at the end?
 
 	On the real thing, there's some kind of weirdness with ADSR when you voice on when attack_rate(raw) = 0x7F; the envelope level register is repeatedly
@@ -64,8 +61,15 @@
 	block-decoding and flag-set program-readable loop address go.
 */
 
+/*
+ Update() isn't called on Read and Writes for performance reasons, it's called with sufficient granularity from the event
+ system, though this will obviously need to change if we ever emulate the SPU with better precision than per-sample(pair).
+*/
+
 
 #include "psx.h"
+#include "cdc.h"
+#include "spu.h"
 
 #include "../clamp.h"
 
@@ -101,8 +105,8 @@ PS_SPU::~PS_SPU()
 
 void PS_SPU::Power(void)
 {
- lastts = 0;
- clock_divider = 0;
+ //lastts = 0;
+ clock_divider = 768;
 
  memset(SPURAM, 0, sizeof(SPURAM));
 
@@ -129,7 +133,7 @@ void PS_SPU::Power(void)
 
   Voices[i].LoopAddr = 0;
 
-  Voices[i].PreLRSample = 0;	// After enveloping, but before L/R volume.  Range of -32768 to 32767
+  Voices[i].PreLRSample = 0;
 
   memset(&Voices[i].ADSR, 0, sizeof(SPU_ADSR));
  }
@@ -162,7 +166,10 @@ void PS_SPU::Power(void)
 
  BlockEnd = 0;
 
- CDDACur = 0;
+ CDAudioCWA = 0;
+
+ memset(CDXA_ResampBuffer, 0, sizeof(CDXA_ResampBuffer));
+ CDXA_CurPhase = 0;
 
  VoiceCWP[0] = 0;
  VoiceCWP[1] = 0;
@@ -175,7 +182,7 @@ void PS_SPU::Power(void)
  memset(RUSB, 0, sizeof(RUSB));
  RUSB_WP = 0;
 
- ReverbCur = 0;
+ ReverbCur = ReverbWA;
 }
 
 static INLINE void CalcVCDelta(const uint8 zs, uint8 speed, bool log_mode, bool dec_mode, int16 Current, int &increment, int &divinco)
@@ -279,6 +286,38 @@ INLINE void SPU_Sweep::WriteVolume(int16 value)
  Current = value;
 }
 
+// output should be readable at -2 and -1
+void PS_SPU::DecodeADPCM(const uint8 *input, int16 *output, const unsigned shift, const unsigned weight)
+{
+ // 5 through 0xF appear to be 0 on the real thing.
+ static const int32 Weights[16][2] =
+ {
+  // s-1    s-2
+  {   0,    0 },
+  {  60,    0 },
+  { 115,  -52 },
+  {  98,  -55 },
+  { 122,  -60 },
+ };
+
+ for(int i = 0; i < 28; i++)
+ {
+  int32 sample;
+
+  sample = (int16)(input[i] << 8);
+  sample >>= shift;
+
+  sample += ((output[i - 1] * Weights[weight][0]) >> 6) + ((output[i - 2] * Weights[weight][1]) >> 6);
+
+  if(sample < -32768)
+   sample = -32768;
+
+  if(sample > 32767)
+   sample = 32767;
+
+  output[i] = sample;
+ }
+}
 
 void PS_SPU::DecodeSamples(SPU_Voice *voice)
 {
@@ -322,7 +361,7 @@ void PS_SPU::DecodeSamples(SPU_Voice *voice)
 
  shift = settings & 0xF;
  weight = (settings >> 4) & 0xF;
- flags = (settings >> 8) & 0xFF;	// FIXME: & 0xF or something else???
+ flags = (settings >> 8) & 0xFF;
 
  for(int i = 0; i < 28; i++)
  {
@@ -358,8 +397,8 @@ void PS_SPU::DecodeSamples(SPU_Voice *voice)
 
   if(TestIRQAddr >= PrevCurAddr && TestIRQAddr < voice->CurAddr)
   {
+   //puts("SPU IRQ");
    IRQ_Assert(IRQ_SPU, true);
-   IRQ_Assert(IRQ_SPU, false);
   }
  }
 
@@ -493,9 +532,8 @@ INLINE void PS_SPU::CheckIRQAddr(uint32 addr)
  {
   if(IRQAddr == addr)
   {
-   puts("SPU IRQ (ALT)");
+   //puts("SPU IRQ (ALT)");
    IRQ_Assert(IRQ_SPU, true);
-   IRQ_Assert(IRQ_SPU, false);
   }
  }
 }
@@ -516,31 +554,71 @@ INLINE uint16 PS_SPU::ReadSPURAM(uint32 addr)
 
 #include "spu_reverb.inc"
 
-void PS_SPU::Update(const pscpu_timestamp_t timestamp)
+INLINE bool PS_SPU::GetCDAudio(int32 &l, int32 &r)
 {
- int32 clocks = timestamp - lastts;
- int32 sample_clocks;
- lastts = timestamp;
+ unsigned freq = CDC->GetCDAudioFreq();
 
- clock_divider += clocks;
- sample_clocks = clock_divider / 768;
- clock_divider -= sample_clocks * 768;
+ if(!freq)
+  return false;
 
- if(!(SPUControl & 0x8000)) // SPU disabled?
+ if(freq == 7 || freq == 14)
  {
-  //PSX_WARNING("[SPU] Disabled?");
-  while(sample_clocks > 0)
+  CDC->GetCDAudio(l, r);
+  if(freq == 14)
+   CDC->GetCDAudio(l, r);
+ }
+ else
+ {
+  const int pi = CDXA_CurPhase * 37;
+  int32 out_tmp[2];
+
+  for(unsigned i = 0; i < 2; i++)
   {
-   assert(IntermediateBufferPos < 4096);
-   IntermediateBuffer[IntermediateBufferPos][0] = 0;
-   IntermediateBuffer[IntermediateBufferPos][1] = 0;
-   IntermediateBufferPos++;
-   sample_clocks--;
+   out_tmp[i] = ((CDXA_ResampBuffer[i][0] * FIR_Table[pi][0]) +
+	         (CDXA_ResampBuffer[i][1] * FIR_Table[pi][1]) +
+	         (CDXA_ResampBuffer[i][2] * FIR_Table[pi][2]) +   
+	         (CDXA_ResampBuffer[i][3] * FIR_Table[pi][3])) >> 15;
+  }
+
+  l = out_tmp[0];
+  r = out_tmp[1];
+
+  CDXA_CurPhase += freq;
+
+  if(CDXA_CurPhase >= 7)
+  {
+   int32 raw[2];
+   CDXA_CurPhase -= 7;
+
+   CDC->GetCDAudio(raw[0], raw[1]);
+
+   for(unsigned i = 0; i < 2; i++)
+   {
+    CDXA_ResampBuffer[i][0] = CDXA_ResampBuffer[i][1];
+    CDXA_ResampBuffer[i][1] = CDXA_ResampBuffer[i][2];
+    CDXA_ResampBuffer[i][2] = CDXA_ResampBuffer[i][3];
+    CDXA_ResampBuffer[i][3] = raw[i];
+   }
   }
-  return;
  }
 
- //printf("SPU Update Start: %d", sample_clocks);
+ return true;
+}
+
+int32 PS_SPU::UpdateFromCDC(int32 clocks)
+//pscpu_timestamp_t PS_SPU::Update(const pscpu_timestamp_t timestamp)
+{
+ //int32 clocks = timestamp - lastts;
+ int32 sample_clocks = 0;
+ //lastts = timestamp;
+
+ clock_divider -= clocks;
+
+ while(clock_divider <= 0)
+ {
+  clock_divider += 768;
+  sample_clocks++;
+ }
 
  while(sample_clocks > 0)
  {
@@ -593,7 +671,7 @@ void PS_SPU::Update(const pscpu_timestamp_t timestamp)
     voice_pvs = ((voice->DecodeBuffer[si + 0] * FIR_Table[pi][0]) +
 	       (voice->DecodeBuffer[si + 1] * FIR_Table[pi][1]) +
 	       (voice->DecodeBuffer[si + 2] * FIR_Table[pi][2]) +   
-	       (voice->DecodeBuffer[si + 3] * FIR_Table[pi][3])) >> 15; //11;
+	       (voice->DecodeBuffer[si + 3] * FIR_Table[pi][3])) >> 15;
    }
 
    voice_pvs = (voice_pvs * (int16)voice->ADSR.EnvLevel) >> 15;
@@ -643,21 +721,15 @@ void PS_SPU::Update(const pscpu_timestamp_t timestamp)
     voice->CurPhase_SD += phase_inc;
    }
 
-   if(VoiceOff & VoiceOn)
-   {
-    PSX_WARNING("[SPU] Simultaneous VoiceOn and VoiceOff --- 0x%08x", VoiceOff & VoiceOn);
-   }
-
-   if(VoiceOff & (1 << voice_num))
+   if(!(SPUControl & 0x8000) || (VoiceOff & (1 << voice_num)))
    {
     if(voice->ADSR.Phase != ADSR_RELEASE)
     {
-     //PSX_WARNING("[SPU] Voice off: %d(%08x %08x)", voice_num, VoiceOff, VoiceOn);
      ReleaseEnvelope(voice);
     }
    }
 
-   if(VoiceOn & (1 << voice_num))
+   if((SPUControl & 0x8000) && (VoiceOn & (1 << voice_num)))
    {
     ResetEnvelope(voice);
 
@@ -677,36 +749,39 @@ void PS_SPU::Update(const pscpu_timestamp_t timestamp)
 
     if(voice_num == 1 || voice_num == 3)
      VoiceCWP[voice_num >> 1] = 0;
-
-    //PSX_WARNING("[SPU] Voice on %d, CurAddr=%08x (%08x %08x)", voice_num, voice->CurAddr, VoiceOff, VoiceOn);
    }
   }
+
   VoiceOff = 0;
   VoiceOn = 0; 
 
   // Get CD-DA
   {
-   int32 cdda_l, cdda_r;
+   int32 cda_raw[2] = { 0 };
 
-   if(CDC->GetCDDA(cdda_l, cdda_r))
+   if(GetCDAudio(cda_raw[0], cda_raw[1]))
    {
-    cdda_l = (cdda_l * CDVol[0]) >> 15;
-    cdda_r = (cdda_r * CDVol[0]) >> 15;
+    int32 cdav[2];
 
-    WriteSPURAM(CDDACur | 0x000, cdda_l);
-    WriteSPURAM(CDDACur | 0x200, cdda_l);
+    for(unsigned i = 0; i < 2; i++)
+     cdav[i] = (cda_raw[i] * CDVol[i]) >> 15;
 
-    CDDACur = (CDDACur + 1) & 0x1FF;
-
-    accum_l += cdda_l;
-    accum_r += cdda_r;
-
-    if(SPUControl & 0x0004)
+    if(SPUControl & 0x0001)
     {
-     accum_fv_l += cdda_l;
-     accum_fv_r += cdda_r;
+     accum_l += cdav[0];
+     accum_r += cdav[1];
+
+     if(SPUControl & 0x0004)	// TODO: Test this bit(and see if it is really dependent on bit0)
+     {
+      accum_fv_l += cdav[0];
+      accum_fv_r += cdav[1];
+     }
     }
    }
+   WriteSPURAM(CDAudioCWA | 0x000, cda_raw[0]);
+   WriteSPURAM(CDAudioCWA | 0x200, cda_raw[1]);
+
+   CDAudioCWA = (CDAudioCWA + 1) & 0x1FF;
   }
 
 
@@ -726,14 +801,11 @@ void PS_SPU::Update(const pscpu_timestamp_t timestamp)
   //accum_fv_l >>= 1;
   //accum_fv_r >>= 1;
 
-  if(!(SPUControl & 0x4000)) // General output muted?
+  if(!(SPUControl & 0x4000)) // General output muted? TODO: Determine if it affects CD audio too; and if not, and we add sub-sample timing accuracy,
+			     // if it's checked for every channel at different times, or just once.
   {
    accum_l = 0;
    accum_r = 0;
-  }
-
-  if(!(SPUControl & 0x80))	// Reverb muted?
-  {
    accum_fv_l = 0;
    accum_fv_r = 0;
   }
@@ -780,7 +852,9 @@ void PS_SPU::Update(const pscpu_timestamp_t timestamp)
    GlobalSweep[lr].Clock();
  }
 
- //puts("SPU Update End");
+ //assert(clock_divider < 768);
+
+ return clock_divider;
 }
 
 void PS_SPU::WriteDMA(uint32 V)
@@ -809,9 +883,7 @@ uint32 PS_SPU::ReadDMA(void)
 
 void PS_SPU::Write(pscpu_timestamp_t timestamp, uint32 A, uint16 V)
 {
- Update(timestamp);
-
- //PSX_WARNING("[SPU] Write: %08x %04x\n", A, V);
+ //PSX_WARNING("[SPU] Write: %08x %04x", A, V);
 
  A &= 0x1FF;
 
@@ -931,15 +1003,16 @@ void PS_SPU::Write(pscpu_timestamp_t timestamp, uint32 A, uint16 V)
 	      break;
 
    case 0x2A: SPUControl = V;
+	      //printf("SPU control write: %04x\n", V);
+	      if(!(V & 0x40))
+	       IRQ_Assert(IRQ_SPU, false);
 	      break;
 
-   case 0x2C: assert(V == 0x4);	//PSX_WARNING("[SPU] Global reg 0x2c set: 0x%04x", V);
+   case 0x2C: PSX_WARNING("[SPU] Global reg 0x2c set: 0x%04x", V);
 	      break;
 
    //case 0x2E: // Read-only?
 
-   // FIXME: volume register sign extension, are sizes 15-bit or 16-bit?
-
    case 0x30: CDVol[0] = V;
 	      break;
 
@@ -963,8 +1036,6 @@ void PS_SPU::Write(pscpu_timestamp_t timestamp, uint32 A, uint16 V)
 
 uint16 PS_SPU::Read(pscpu_timestamp_t timestamp, uint32 A)
 {
- Update(timestamp);
-
  A &= 0x1FF;
 
  PSX_DBGINFO("[SPU] Read: %08x", A);
@@ -1005,6 +1076,7 @@ uint16 PS_SPU::Read(pscpu_timestamp_t timestamp, uint32 A)
    // FIXME: What are these two for?
    case 0x2e:
    case 0x3C:
+	//PSX_WARNING("[SPU] Read Unknown: %08x", A);
 	return(0);
 
    case 0x38:
@@ -1030,7 +1102,10 @@ void PS_SPU::StartFrame(double rate, uint32 quality)
    resampler = NULL;
   }
 
-  resampler = speex_resampler_init(2, 44100, (int)rate, quality, &err);
+  if((int)rate && (int)rate != 44100)
+  {
+   resampler = speex_resampler_init(2, 44100, (int)rate, quality, &err);
+  }
 
   last_rate = (int)rate;
   last_quality = quality;
@@ -1040,7 +1115,7 @@ void PS_SPU::StartFrame(double rate, uint32 quality)
 
 int32 PS_SPU::EndFrame(int16 *SoundBuf)
 {
- lastts = 0;
+ //lastts = 0;
 
  if(last_rate == 44100)
  {
@@ -1051,13 +1126,13 @@ int32 PS_SPU::EndFrame(int16 *SoundBuf)
 
   return(ret);
  }
- else
+ else if(resampler)
  {
   spx_uint32_t in_len;
   spx_uint32_t out_len;
 
   in_len = IntermediateBufferPos;
-  out_len = 2048;	// FIXME, real size.
+  out_len = 524288; //8192;	// FIXME, real size.
 
   speex_resampler_process_interleaved_int(resampler, (const spx_int16_t *)IntermediateBuffer, &in_len, (spx_int16_t *)SoundBuf, &out_len);
 
@@ -1070,6 +1145,11 @@ int32 PS_SPU::EndFrame(int16 *SoundBuf)
 
   return(out_len);
  }
+ else
+ {
+  IntermediateBufferPos = 0;
+  return 0;
+ }
 }
 
 
@@ -1123,6 +1203,10 @@ uint32 PS_SPU::GetRegister(unsigned int which, char *special, const uint32 speci
    case GSREG_V0_LOOP_ADDR:
 	ret = Voices[v].LoopAddr;
 	break;
+
+   case GSREG_V0_READ_ADDR:
+	ret = Voices[v].CurAddr;
+	break;
   }
  }
  else switch(which)
@@ -1187,6 +1271,19 @@ uint32 PS_SPU::GetRegister(unsigned int which, char *special, const uint32 speci
 	ret = ReverbWA >> 2;
 	break;
 
+  case GSREG_VOICEON:
+	ret = VoiceOn;
+	break;
+
+  case GSREG_VOICEOFF:
+	ret = VoiceOff;
+	break;
+
+  case GSREG_BLOCKEND:
+	ret = BlockEnd;
+	break;
+
+
   case GSREG_FB_SRC_A ... GSREG_IN_COEF_R:
 	ret = ReverbRegs[which - GSREG_FB_SRC_A] & 0xFFFF;
 	break;
@@ -1260,6 +1357,23 @@ void PS_SPU::SetRegister(unsigned int which, uint32 value)
 	IRQAddr = value & 0x3FFFC;
 	break;
 
+  //
+  // REVERB_WA
+  //
+
+  case GSREG_VOICEON:
+        VoiceOn = value & 0xFFFFFF;
+        break;
+
+  case GSREG_VOICEOFF:
+        VoiceOff = value & 0xFFFFFF;
+        break;
+
+  case GSREG_BLOCKEND:
+        BlockEnd = value & 0xFFFFFF;
+        break;
+
+
  }
 }
 
