@@ -28,8 +28,9 @@ uint16 TIMER_Read(const pscpu_timestamp_t timestamp, uint32 A);
 
 void TIMER_AddDotClocks(uint32 count);
 void TIMER_ClockHRetrace(void);
+void TIMER_SetVSync(bool status);	// TODO
 
-void TIMER_Update(const pscpu_timestamp_t);
+pscpu_timestamp_t TIMER_Update(const pscpu_timestamp_t);
 void TIMER_ResetTS(void);
 
 void TIMER_Power(void);
