@@ -202,7 +202,9 @@ static bool ValidateSetting(const char *value, const MDFNSetting *setting)
     found = true;
     break;
    }
-   valid_string_list = valid_string_list + std::string(enum_list->string) + (enum_list[1].string ? " " : "");
+
+   if(enum_list->description)	// Don't list out undocumented and deprecated values.
+    valid_string_list = valid_string_list + (enum_list == setting->enum_list ? "" : " ") + std::string(enum_list->string);
 
    enum_list++;
   }
