@@ -55,7 +55,8 @@ static int16 MouseYLatch[2];
 static uint8 MouseBLatch[2];
 
 static uint8 *CustomColorMap = NULL;
-static uint32 ColorMap[32768];
+//static uint32 ColorMap[32768];
+static std::vector<uint32> ColorMap;
 
 static bool LoadCPalette(const char *syspalname, uint8 **ptr, uint32 num_entries)
 {
@@ -479,6 +480,8 @@ static int Load(const char *name, MDFNFILE *fp)
 
  MDFNMP_AddRAM(131072, 0x7E << 16, SNES::memory::wram.data());
 
+ ColorMap.resize(32768);
+
  if(!LoadCPalette(NULL, &CustomColorMap, 32768))
   return(0);	// FIXME: cleanup
 
@@ -620,7 +623,7 @@ struct StrToBSIT_t
  const int id;
 };
 
-static StrToBSIT_t StrToBSIT[] =
+static const StrToBSIT_t StrToBSIT[] =
 {
  { "none",   	SNES::Input::DeviceNone },
  { "gamepad",   SNES::Input::DeviceJoypad },
@@ -639,7 +642,7 @@ static void SetInput(int port, const char *type, void *ptr)
 
  if(port < 2)
  {
-  StrToBSIT_t *sb = StrToBSIT;
+  const StrToBSIT_t *sb = StrToBSIT;
   int id = -1;
 
   if(MultitapEnabled[port] && !strcmp(type, "gamepad"))
@@ -716,7 +719,7 @@ static const InputDeviceInputInfoStruct GamepadIDII[] =
  { "r", "Right Shoulder", 11, IDIT_BUTTON, NULL },
 };
 
-const InputDeviceInputInfoStruct MouseIDII[0x4] =
+static const InputDeviceInputInfoStruct MouseIDII[0x4] =
 {
  { "x_axis", "X Axis", -1, IDIT_X_AXIS_REL },
  { "y_axis", "Y Axis", -1, IDIT_Y_AXIS_REL },
@@ -788,7 +791,7 @@ static InputInfoStruct SNESInputInfo =
  PortInfo
 };
 
-static MDFNSetting SNESSettings[] =
+static const MDFNSetting SNESSettings[] =
 {
  { "snes.input.port1.multitap", MDFNSF_EMU_STATE | MDFNSF_UNTRUSTED_SAFE, gettext_noop("Enable multitap on SNES port 1."), NULL, MDFNST_BOOL, "0", NULL, NULL },
  { "snes.input.port2.multitap", MDFNSF_EMU_STATE | MDFNSF_UNTRUSTED_SAFE, gettext_noop("Enable multitap on SNES port 2."), NULL, MDFNST_BOOL, "0", NULL, NULL },
