@@ -367,7 +367,7 @@ static int ReadStateChunk(StateMem *st, SFORMAT *sf, int size, int data_only)
  else
  {
   SFMap_t sfmap;
-  SFMap_t sfmap_found;	// Used for identify variables that are missing in the save state.
+  SFMap_t sfmap_found;	// Used for identifying variables that are missing in the save state.
 
   MakeSFMap(sf, sfmap);
 
